@@ -130,6 +130,7 @@ void addOptimizationPasses(legacy::PassManagerBase *PM, int opt_level,
             PM->add(createLowerExcHandlersPass());
             PM->add(createGCInvariantVerifierPass(false));
             PM->add(createLateLowerGCFramePass());
+            PM->add(createFinalLowerGCPass());
             PM->add(createLowerPTLSPass(dump_native));
         }
         PM->add(createLowerSimdLoopPass());        // Annotate loop marked with "loopinfo" as LLVM parallel loop
@@ -245,6 +246,7 @@ void addOptimizationPasses(legacy::PassManagerBase *PM, int opt_level,
         PM->add(createLowerExcHandlersPass());
         PM->add(createGCInvariantVerifierPass(false));
         PM->add(createLateLowerGCFramePass());
+        PM->add(createFinalLowerGCPass());
         // Remove dead use of ptls
         PM->add(createDeadCodeEliminationPass());
         PM->add(createLowerPTLSPass(dump_native));
@@ -319,7 +321,7 @@ void JuliaOJIT::DebugObjectRegistrar::registerObject(RTDyldObjHandleT H, const O
                                                        std::move(NewBuffer));
     }
     else {
-        JIT.NotifyFinalizer(*(SavedObject.getBinary()), *LO);
+        JIT.NotifyFinalizer(H, *(SavedObject.getBinary()), *LO);
     }
 
     SavedObjects.push_back(std::move(SavedObject));
@@ -437,9 +439,9 @@ JuliaOJIT::JuliaOJIT(TargetMachine &TM)
     // Make sure SectionMemoryManager::getSymbolAddressInProcess can resolve
     // symbols in the program as well. The nullptr argument to the function
     // tells DynamicLibrary to load the program, not a library.
-    std::string *ErrorStr = nullptr;
-    if (sys::DynamicLibrary::LoadLibraryPermanently(nullptr, ErrorStr))
-        report_fatal_error("FATAL: unable to dlopen self\n" + *ErrorStr);
+    std::string ErrorStr;
+    if (sys::DynamicLibrary::LoadLibraryPermanently(nullptr, &ErrorStr))
+        report_fatal_error("FATAL: unable to dlopen self\n" + ErrorStr);
 }
 
 void JuliaOJIT::addGlobalMapping(StringRef Name, uint64_t Addr)
@@ -579,11 +581,17 @@ void JuliaOJIT::RegisterJITEventListener(JITEventListener *L)
     EventListeners.push_back(L);
 }
 
-void JuliaOJIT::NotifyFinalizer(const object::ObjectFile &Obj,
+void JuliaOJIT::NotifyFinalizer(RTDyldObjHandleT Key,
+                                const object::ObjectFile &Obj,
                                 const RuntimeDyld::LoadedObjectInfo &LoadedObjectInfo)
 {
     for (auto &Listener : EventListeners)
+#if JL_LLVM_VERSION >= 80000
+        Listener->notifyObjectLoaded(Key, Obj, LoadedObjectInfo);
+#else
         Listener->NotifyObjectEmitted(Obj, LoadedObjectInfo);
+    (void)Key;
+#endif
 }
 
 const DataLayout& JuliaOJIT::getDataLayout() const
@@ -888,7 +896,7 @@ static std::map<void*, jl_value_llvm> jl_value_to_llvm;
 //
 // then add a global mapping to the current value (usually from calloc'd space)
 // to the execution engine to make it valid for the current session (with the current value)
-void* jl_emit_and_add_to_shadow(GlobalVariable *gv, void *gvarinit)
+void** jl_emit_and_add_to_shadow(GlobalVariable *gv, void *gvarinit)
 {
     PointerType *T = cast<PointerType>(gv->getType()->getElementType()); // pointer is the only supported type here
 
@@ -911,7 +919,7 @@ void* jl_emit_and_add_to_shadow(GlobalVariable *gv, void *gvarinit)
     }
 
     // make the pointer valid for this session
-    void *slot = calloc(1, sizeof(void*));
+    void **slot = (void**)calloc(1, sizeof(void*));
     jl_ExecutionEngine->addGlobalMapping(gv, slot);
     return slot;
 }
@@ -1165,7 +1173,7 @@ GlobalVariable *jl_get_global_for(const char *cname, void *addr, Module *M)
     GlobalVariable *gv = new GlobalVariable(*M, T_pjlvalue,
                            false, GlobalVariable::ExternalLinkage,
                            NULL, gvname.str());
-    *(void**)jl_emit_and_add_to_shadow(gv, addr) = addr;
+    *jl_emit_and_add_to_shadow(gv, addr) = addr;
     return gv;
 }
 
