@@ -23,8 +23,8 @@ jl_datatype_t *jl_any_type;
 jl_unionall_t *jl_type_type;
 jl_typename_t *jl_type_typename;
 jl_methtable_t *jl_type_type_mt;
+jl_methtable_t *jl_nonfunction_mt;
 jl_datatype_t *jl_typename_type;
-jl_datatype_t *jl_sym_type;
 jl_datatype_t *jl_symbol_type;
 jl_datatype_t *jl_ssavalue_type;
 jl_datatype_t *jl_abstractslot_type;
@@ -119,6 +119,7 @@ jl_datatype_t *jl_void_type;
 jl_datatype_t *jl_voidpointer_type;
 jl_typename_t *jl_namedtuple_typename;
 jl_unionall_t *jl_namedtuple_type;
+jl_datatype_t *jl_task_type;
 jl_value_t *jl_an_empty_vec_any;
 jl_value_t *jl_an_empty_string;
 jl_value_t *jl_stackovf_exception;
@@ -126,19 +127,15 @@ jl_value_t *jl_stackovf_exception;
 jl_value_t *jl_segv_exception;
 #endif
 JL_DLLEXPORT jl_value_t *jl_diverror_exception;
-JL_DLLEXPORT jl_value_t *jl_domain_exception;
-JL_DLLEXPORT jl_value_t *jl_overflow_exception;
 JL_DLLEXPORT jl_value_t *jl_undefref_exception;
 jl_value_t *jl_interrupt_exception;
 jl_datatype_t *jl_boundserror_type;
 jl_value_t *jl_memory_exception;
 jl_value_t *jl_readonlymemory_exception;
 
-jl_cgparams_t jl_default_cgparams = {1, 1, 1, 1, 0, NULL, NULL, NULL, NULL, NULL};
-
 // --- type properties and predicates ---
 
-static int typeenv_has(jl_typeenv_t *env, jl_tvar_t *v)
+static int typeenv_has(jl_typeenv_t *env, jl_tvar_t *v) JL_NOTSAFEPOINT
 {
     while (env != NULL) {
         if (env->var == v)
@@ -148,7 +145,7 @@ static int typeenv_has(jl_typeenv_t *env, jl_tvar_t *v)
     return 0;
 }
 
-static int has_free_typevars(jl_value_t *v, jl_typeenv_t *env)
+static int has_free_typevars(jl_value_t *v, jl_typeenv_t *env) JL_NOTSAFEPOINT
 {
     if (jl_typeis(v, jl_tvar_type)) {
         return !typeenv_has(env, (jl_tvar_t*)v);
@@ -177,7 +174,7 @@ static int has_free_typevars(jl_value_t *v, jl_typeenv_t *env)
     return 0;
 }
 
-JL_DLLEXPORT int jl_has_free_typevars(jl_value_t *v)
+JL_DLLEXPORT int jl_has_free_typevars(jl_value_t *v) JL_NOTSAFEPOINT
 {
     return has_free_typevars(v, NULL);
 }
@@ -218,7 +215,7 @@ JL_DLLEXPORT jl_array_t *jl_find_free_typevars(jl_value_t *v)
 }
 
 // test whether a type has vars bound by the given environment
-static int jl_has_bound_typevars(jl_value_t *v, jl_typeenv_t *env)
+static int jl_has_bound_typevars(jl_value_t *v, jl_typeenv_t *env) JL_NOTSAFEPOINT
 {
     if (jl_typeis(v, jl_tvar_type))
         return typeenv_has(env, (jl_tvar_t*)v);
@@ -252,7 +249,7 @@ static int jl_has_bound_typevars(jl_value_t *v, jl_typeenv_t *env)
     return 0;
 }
 
-JL_DLLEXPORT int jl_has_typevar(jl_value_t *t, jl_tvar_t *v)
+JL_DLLEXPORT int jl_has_typevar(jl_value_t *t, jl_tvar_t *v) JL_NOTSAFEPOINT
 {
     jl_typeenv_t env = { v, NULL, NULL };
     return jl_has_bound_typevars(t, &env);
@@ -640,6 +637,8 @@ void jl_sort_types(jl_value_t **types, size_t length)
 static int typekey_eq(jl_datatype_t *tt, jl_value_t **key, size_t n)
 {
     size_t j;
+    // TOOD: This shouldn't be necessary
+    JL_GC_PROMISE_ROOTED(tt);
     size_t tnp = jl_nparams(tt);
     if (n != tnp) return 0;
     if (tt->name == jl_type_typename) {
@@ -690,7 +689,8 @@ static ssize_t lookup_type_idx(jl_typename_t *tn, jl_value_t **key, size_t n, in
           need to allow sequences of typekey_compare-equal types in the ordered cache.
         */
         while (hi < cl && typekey_compare(data[hi], key, n) == 0) {
-            if (typekey_eq(data[hi], key, n))
+            jl_datatype_t *tt = data[hi];
+            if (typekey_eq(tt, key, n))
                 return hi;
             hi++;
         }
@@ -935,7 +935,7 @@ JL_EXTENSION struct _jl_typestack_t {
 };
 
 static jl_value_t *inst_type_w_(jl_value_t *t, jl_typeenv_t *env, jl_typestack_t *stack, int check);
-static jl_svec_t *inst_all(jl_svec_t *p, jl_typeenv_t *env, jl_typestack_t *stack, int check);
+static jl_svec_t *inst_ftypes(jl_svec_t *p, jl_typeenv_t *env, jl_typestack_t *stack);
 
 JL_DLLEXPORT jl_value_t *jl_instantiate_unionall(jl_unionall_t *u, jl_value_t *p)
 {
@@ -1132,8 +1132,17 @@ static jl_value_t *inst_datatype_inner(jl_datatype_t *dt, jl_svec_t *p, jl_value
 
     if (!istuple) {
         if (jl_is_vararg_type((jl_value_t*)dt) && ntp == 2) {
-            if (!jl_is_long(iparams[1]) && !jl_is_typevar(iparams[1])) {
-                jl_type_error_rt("Vararg", "count", (jl_value_t*)jl_long_type, iparams[1]);
+            jl_value_t *lenparam = iparams[1];
+            if (jl_is_typevar(lenparam)) {
+                jl_tvar_t *N = (jl_tvar_t*)lenparam;
+                if (!(N->lb == jl_bottom_type && N->ub == (jl_value_t*)jl_any_type))
+                    jl_error("TypeVar in Vararg length must have bounds Union{} and Any");
+            }
+            else if (!jl_is_long(lenparam)) {
+                jl_type_error_rt("Vararg", "count", (jl_value_t*)jl_long_type, lenparam);
+            }
+            else if (jl_unbox_long(lenparam) < 0) {
+                jl_errorf("Vararg length is negative: %zd", jl_unbox_long(lenparam));
             }
         }
         // check parameters against bounds in type definition
@@ -1175,8 +1184,7 @@ static jl_value_t *inst_datatype_inner(jl_datatype_t *dt, jl_svec_t *p, jl_value
         }
         if (jl_is_long(va1)) {
             ssize_t nt = jl_unbox_long(va1);
-            if (nt < 0)
-                jl_errorf("apply_type: Vararg length N is negative: %zd", nt);
+            assert(nt >= 0);
             if (nt == 0 || !jl_has_free_typevars(va0)) {
                 if (cacheable) JL_UNLOCK(&typecache_lock); // Might GC
                 if (ntp == 1) {
@@ -1278,6 +1286,9 @@ static jl_value_t *inst_datatype_inner(jl_datatype_t *dt, jl_svec_t *p, jl_value
         ndt->super = (jl_datatype_t*)inst_type_w_((jl_value_t*)dt->super, env, stack, 1);
         jl_gc_wb(ndt, ndt->super);
     }
+    if (cacheable) {
+        jl_cache_type_(ndt);
+    }
     jl_svec_t *ftypes = dt->types;
     if (ftypes == NULL)
         ftypes = primarydt->types;
@@ -1287,12 +1298,18 @@ static jl_value_t *inst_datatype_inner(jl_datatype_t *dt, jl_svec_t *p, jl_value
         assert((inside_typedef || primarydt->super) && !istuple && !isnamedtuple);
         arraylist_push(&partial_inst, ndt);
     }
-    else {
-        assert(ftypes != jl_emptysvec || jl_field_names(ndt) == jl_emptysvec || isnamedtuple);
+    else if (!isnamedtuple && !istuple) {
+        assert(ftypes != jl_emptysvec || jl_field_names(ndt) == jl_emptysvec);
         assert(ftypes == jl_emptysvec || !ndt->abstract);
-        if (!istuple && !isnamedtuple) {
+        if (ftypes == jl_emptysvec) {
+            ndt->types = ftypes;
+        }
+        else if (cacheable) {
             // recursively instantiate the types of the fields
-            ndt->types = inst_all(ftypes, env, stack, 1);
+            if (dt->types == NULL)
+                ndt->types = jl_compute_fieldtypes(ndt);
+            else
+                ndt->types = inst_ftypes(ftypes, env, stack);
             jl_gc_wb(ndt, ndt->types);
         }
     }
@@ -1313,7 +1330,6 @@ static jl_value_t *inst_datatype_inner(jl_datatype_t *dt, jl_svec_t *p, jl_value
         ndt->ninitialized = dt->ninitialized;
 
     if (cacheable) {
-        jl_cache_type_(ndt);
         JL_UNLOCK(&typecache_lock); // Might GC
     }
 
@@ -1352,14 +1368,21 @@ jl_datatype_t *jl_inst_concrete_tupletype_v(jl_value_t **p, size_t np)
     return (jl_datatype_t*)inst_datatype_inner(jl_anytuple_type, NULL, p, np, 1, NULL, NULL);
 }
 
-static jl_svec_t *inst_all(jl_svec_t *p, jl_typeenv_t *env, jl_typestack_t *stack, int check)
+static jl_svec_t *inst_ftypes(jl_svec_t *p, jl_typeenv_t *env, jl_typestack_t *stack)
 {
     size_t i;
     size_t lp = jl_svec_len(p);
     jl_svec_t *np = jl_alloc_svec(lp);
     JL_GC_PUSH1(&np);
-    for(i=0; i < lp; i++) {
-        jl_svecset(np, i, (jl_value_t*)inst_type_w_(jl_svecref(p,i), env, stack, check));
+    for (i = 0; i < lp; i++) {
+        jl_value_t *pi = jl_svecref(p, i);
+        JL_TRY {
+            pi = inst_type_w_(pi, env, stack, 1);
+        }
+        JL_CATCH {
+            pi = jl_bottom_type;
+        }
+        jl_svecset(np, i, pi);
     }
     JL_GC_POP();
     return np;
@@ -1410,7 +1433,7 @@ static jl_value_t *inst_tuple_w_(jl_value_t *t, jl_typeenv_t *env, jl_typestack_
     int i;
     for (i = 0; i < ntp; i++) {
         jl_value_t *elt = jl_svecref(tp, i);
-        jl_value_t *pi = (jl_value_t*)inst_type_w_(elt, env, stack, 0);
+        jl_value_t *pi = inst_type_w_(elt, env, stack, 0);
         iparams[i] = pi;
         if (ip_heap)
             jl_gc_wb(ip_heap, pi);
@@ -1506,18 +1529,18 @@ static jl_value_t *inst_type_w_(jl_value_t *t, jl_typeenv_t *env, jl_typestack_t
     return t;
 }
 
-jl_value_t *instantiate_with(jl_value_t *t, jl_value_t **env, size_t n, jl_typeenv_t *te, jl_typestack_t *stack)
+static jl_value_t *instantiate_with(jl_value_t *t, jl_value_t **env, size_t n, jl_typeenv_t *te)
 {
     if (n > 0) {
         jl_typeenv_t en = { (jl_tvar_t*)env[0], env[1], te };
-        return instantiate_with(t, &env[2], n-1, &en, stack);
+        return instantiate_with(t, &env[2], n-1, &en );
     }
-    return inst_type_w_(t, te, stack, 1);
+    return inst_type_w_(t, te, NULL, 1);
 }
 
 jl_value_t *jl_instantiate_type_with(jl_value_t *t, jl_value_t **env, size_t n)
 {
-    return instantiate_with(t, env, n, NULL, NULL);
+    return instantiate_with(t, env, n, NULL);
 }
 
 static jl_value_t *_jl_instantiate_type_in_env(jl_value_t *ty, jl_unionall_t *env, jl_value_t **vals, jl_typeenv_t *prev, jl_typestack_t *stack)
@@ -1563,6 +1586,29 @@ jl_value_t *jl_wrap_vararg(jl_value_t *t, jl_value_t *n)
     return vn;
 }
 
+JL_DLLEXPORT jl_svec_t *jl_compute_fieldtypes(jl_datatype_t *st JL_PROPAGATES_ROOT)
+{
+    assert(st->name != jl_namedtuple_typename && st->name != jl_tuple_typename);
+    jl_datatype_t *wt = (jl_datatype_t*)jl_unwrap_unionall(st->name->wrapper);
+    size_t i, n = jl_svec_len(wt->parameters);
+    assert(n > 0 && "expected empty case to be handled during construction");
+    //if (n == 0)
+    //    return ((st->types = jl_emptysvec));
+    jl_typeenv_t *env = (jl_typeenv_t*)alloca(n * sizeof(jl_typeenv_t));
+    for (i = 0; i < n; i++) {
+        env[i].var = (jl_tvar_t*)jl_svecref(wt->parameters, i);
+        env[i].val = jl_svecref(st->parameters, i);
+        env[i].prev = i == 0 ? NULL : &env[i - 1];
+    }
+    jl_typestack_t top;
+    top.tt = st;
+    top.prev = NULL;
+    st->types = inst_ftypes(wt->types, &env[n - 1], &top);
+    jl_gc_wb(st, st->types);
+    return st->types;
+}
+
+
 void jl_reinstantiate_inner_types(jl_datatype_t *t) // can throw!
 {
     inside_typedef = 0;
@@ -1576,19 +1622,20 @@ void jl_reinstantiate_inner_types(jl_datatype_t *t) // can throw!
         return;
     }
 
-    jl_value_t **env = (jl_value_t**)alloca(n * 2 * sizeof(void*));
+    jl_typeenv_t *env = (jl_typeenv_t*)alloca(n * sizeof(jl_typeenv_t));
     for (i = 0; i < n; i++) {
-        env[i * 2] = jl_svecref(t->parameters, i);
-        env[i * 2 + 1] = NULL;
+        env[i].var = (jl_tvar_t*)jl_svecref(t->parameters, i);
+        env[i].val = NULL;
+        env[i].prev = i == 0 ? NULL : &env[i - 1];
     }
 
     for (j = 0; j < partial_inst.len; j++) {
         jl_datatype_t *ndt = (jl_datatype_t*)partial_inst.items[j];
         assert(jl_unwrap_unionall(ndt->name->wrapper) == (jl_value_t*)t);
         for (i = 0; i < n; i++)
-            env[i * 2 + 1] = jl_svecref(ndt->parameters, i);
+            env[i].val = jl_svecref(ndt->parameters, i);
 
-        ndt->super = (jl_datatype_t*)instantiate_with((jl_value_t*)t->super, env, n, NULL, &top);
+        ndt->super = (jl_datatype_t*)inst_type_w_((jl_value_t*)t->super, env, &top, 1);
         jl_gc_wb(ndt, ndt->super);
     }
 
@@ -1596,18 +1643,10 @@ void jl_reinstantiate_inner_types(jl_datatype_t *t) // can throw!
         for (j = 0; j < partial_inst.len; j++) {
             jl_datatype_t *ndt = (jl_datatype_t*)partial_inst.items[j];
             for (i = 0; i < n; i++)
-                env[i * 2 + 1] = jl_svecref(ndt->parameters, i);
-
-            int k;
+                env[i].val = jl_svecref(ndt->parameters, i);
             assert(ndt->types == NULL);
-            jl_svec_t *ft = jl_alloc_svec(jl_svec_len(t->types));
-            JL_GC_PUSH1(&ft);
-            for (k=0; k < jl_svec_len(t->types); k++) {
-                jl_svecset(ft, k, instantiate_with(jl_svecref(t->types,k), env, n, NULL, &top));
-            }
-            ndt->types = ft;
+            ndt->types = inst_ftypes(t->types, env, &top);
             jl_gc_wb(ndt, ndt->types);
-            JL_GC_POP();
             if (ndt->uid) { // cacheable
                 jl_compute_field_offsets(ndt);
             }
@@ -1628,40 +1667,36 @@ void jl_reset_instantiate_inner_types(jl_datatype_t *t)
 
 // initialization -------------------------------------------------------------
 
+extern void jl_init_int32_int64_cache(void);
+
 static jl_tvar_t *tvar(const char *name)
 {
     return jl_new_typevar(jl_symbol(name), (jl_value_t*)jl_bottom_type,
                           (jl_value_t*)jl_any_type);
 }
 
-extern void jl_init_int32_int64_cache(void);
-
 void jl_init_types(void) JL_GC_DISABLED
 {
     jl_module_t *core = NULL; // will need to be assigned later
 
-    arraylist_new(&partial_inst, 0);
     // create base objects
     jl_datatype_type = jl_new_uninitialized_datatype();
     jl_set_typeof(jl_datatype_type, jl_datatype_type);
     jl_typename_type = jl_new_uninitialized_datatype();
-    jl_sym_type = jl_new_uninitialized_datatype();
-    jl_symbol_type = jl_sym_type;
+    jl_symbol_type = jl_new_uninitialized_datatype();
     jl_simplevector_type = jl_new_uninitialized_datatype();
     jl_methtable_type = jl_new_uninitialized_datatype();
     jl_nothing = jl_gc_permobj(0, NULL);
 
-    jl_default_cgparams.module_setup = jl_nothing;
-    jl_default_cgparams.module_activation = jl_nothing;
-    jl_default_cgparams.raise_exception = jl_nothing;
-    jl_default_cgparams.emit_function = jl_nothing;
-    jl_default_cgparams.emitted_function = jl_nothing;
-
     jl_emptysvec = (jl_svec_t*)jl_gc_permobj(sizeof(void*), jl_simplevector_type);
     jl_svec_set_len_unsafe(jl_emptysvec, 0);
 
     jl_any_type = (jl_datatype_t*)jl_new_abstracttype((jl_value_t*)jl_symbol("Any"), core, NULL, jl_emptysvec);
     jl_any_type->super = jl_any_type;
+    jl_nonfunction_mt = jl_any_type->name->mt;
+    jl_any_type->name->mt = NULL;
+    jl_nonfunction_mt->offs = 0;
+
     jl_type_type = (jl_unionall_t*)jl_new_abstracttype((jl_value_t*)jl_symbol("Type"), core, jl_any_type, jl_emptysvec);
     jl_type_typename = ((jl_datatype_t*)jl_type_type)->name;
     jl_type_type_mt = jl_new_method_table(jl_type_typename->name, core);
@@ -1672,7 +1707,7 @@ void jl_init_types(void) JL_GC_DISABLED
     jl_datatype_type->name->wrapper = (jl_value_t*)jl_datatype_type;
     jl_datatype_type->super = (jl_datatype_t*)jl_type_type;
     jl_datatype_type->parameters = jl_emptysvec;
-    jl_datatype_type->name->names = jl_perm_symsvec(20,
+    jl_datatype_type->name->names = jl_perm_symsvec(21,
                                                     "name",
                                                     "super",
                                                     "parameters",
@@ -1691,9 +1726,10 @@ void jl_init_types(void) JL_GC_DISABLED
                                                     "isbitstype",
                                                     "zeroinit",
                                                     "isinlinealloc",
+                                                    "has_concrete_subtype",
                                                     "llvm::StructType",
                                                     "llvm::DIType");
-    jl_datatype_type->types = jl_svec(20,
+    jl_datatype_type->types = jl_svec(21,
                                       jl_typename_type,
                                       jl_datatype_type,
                                       jl_simplevector_type,
@@ -1702,7 +1738,7 @@ void jl_init_types(void) JL_GC_DISABLED
                                       jl_any_type, jl_any_type, jl_any_type, jl_any_type,
                                       jl_any_type, jl_any_type, jl_any_type, jl_any_type,
                                       jl_any_type, jl_any_type, jl_any_type, jl_any_type,
-                                      jl_any_type, jl_any_type);
+                                      jl_any_type, jl_any_type, jl_any_type);
     jl_datatype_type->instance = NULL;
     jl_datatype_type->uid = jl_assign_type_uid();
     jl_datatype_type->struct_decl = NULL;
@@ -1711,19 +1747,19 @@ void jl_init_types(void) JL_GC_DISABLED
     // NOTE: types should not really be mutable, but the instance and
     // struct_decl fields are basically caches, which are mutated.
     jl_datatype_type->mutabl = 1;
-    jl_datatype_type->ninitialized = 4;
+    jl_datatype_type->ninitialized = 3;
     jl_precompute_memoized_dt(jl_datatype_type);
 
     jl_typename_type->name = jl_new_typename_in(jl_symbol("TypeName"), core);
     jl_typename_type->name->wrapper = (jl_value_t*)jl_typename_type;
-    jl_typename_type->name->mt = jl_new_method_table(jl_typename_type->name->name, core);
+    jl_typename_type->name->mt = jl_nonfunction_mt;
     jl_typename_type->super = jl_any_type;
     jl_typename_type->parameters = jl_emptysvec;
     jl_typename_type->name->names = jl_perm_symsvec(8, "name", "module",
                                                     "names", "wrapper",
                                                     "cache", "linearcache",
                                                     "hash", "mt");
-    jl_typename_type->types = jl_svec(8, jl_sym_type, jl_any_type, jl_simplevector_type,
+    jl_typename_type->types = jl_svec(8, jl_symbol_type, jl_any_type, jl_simplevector_type,
                                       jl_type_type, jl_simplevector_type, jl_simplevector_type,
                                       jl_any_type, jl_any_type);
     jl_typename_type->uid = jl_assign_type_uid();
@@ -1737,17 +1773,17 @@ void jl_init_types(void) JL_GC_DISABLED
 
     jl_methtable_type->name = jl_new_typename_in(jl_symbol("MethodTable"), core);
     jl_methtable_type->name->wrapper = (jl_value_t*)jl_methtable_type;
-    jl_methtable_type->name->mt = jl_new_method_table(jl_methtable_type->name->name, core);
+    jl_methtable_type->name->mt = jl_nonfunction_mt;
     jl_methtable_type->super = jl_any_type;
     jl_methtable_type->parameters = jl_emptysvec;
-    jl_methtable_type->name->names = jl_perm_symsvec(10, "name", "defs",
+    jl_methtable_type->name->names = jl_perm_symsvec(11, "name", "defs",
                                                      "cache", "max_args",
                                                      "kwsorter", "module",
-                                                     "backedges", "", "", "offs");
-    jl_methtable_type->types = jl_svec(10, jl_sym_type, jl_any_type, jl_any_type, jl_any_type/*jl_long*/,
+                                                     "backedges", "", "", "offs", "");
+    jl_methtable_type->types = jl_svec(11, jl_symbol_type, jl_any_type, jl_any_type, jl_any_type/*jl_long*/,
                                        jl_any_type, jl_any_type/*module*/,
                                        jl_any_type/*any vector*/, jl_any_type/*long*/, jl_any_type/*int32*/,
-                                       jl_any_type/*uint8*/);
+                                       jl_any_type/*uint8*/, jl_any_type/*uint8*/);
     jl_methtable_type->uid = jl_assign_type_uid();
     jl_methtable_type->instance = NULL;
     jl_methtable_type->struct_decl = NULL;
@@ -1757,26 +1793,26 @@ void jl_init_types(void) JL_GC_DISABLED
     jl_methtable_type->ninitialized = 4;
     jl_precompute_memoized_dt(jl_methtable_type);
 
-    jl_sym_type->name = jl_new_typename_in(jl_symbol("Symbol"), core);
-    jl_sym_type->name->wrapper = (jl_value_t*)jl_sym_type;
-    jl_sym_type->name->mt = jl_new_method_table(jl_sym_type->name->name, core);
-    jl_sym_type->super = jl_any_type;
-    jl_sym_type->parameters = jl_emptysvec;
-    jl_sym_type->name->names = jl_emptysvec;
-    jl_sym_type->types = jl_emptysvec;
-    jl_sym_type->instance = NULL;
-    jl_sym_type->uid = jl_assign_type_uid();
-    jl_sym_type->struct_decl = NULL;
-    jl_sym_type->ditype = NULL;
-    jl_sym_type->size = 0;
-    jl_sym_type->abstract = 0;
-    jl_sym_type->mutabl = 1;
-    jl_sym_type->ninitialized = 0;
-    jl_precompute_memoized_dt(jl_sym_type);
+    jl_symbol_type->name = jl_new_typename_in(jl_symbol("Symbol"), core);
+    jl_symbol_type->name->wrapper = (jl_value_t*)jl_symbol_type;
+    jl_symbol_type->name->mt = jl_nonfunction_mt;
+    jl_symbol_type->super = jl_any_type;
+    jl_symbol_type->parameters = jl_emptysvec;
+    jl_symbol_type->name->names = jl_emptysvec;
+    jl_symbol_type->types = jl_emptysvec;
+    jl_symbol_type->instance = NULL;
+    jl_symbol_type->uid = jl_assign_type_uid();
+    jl_symbol_type->struct_decl = NULL;
+    jl_symbol_type->ditype = NULL;
+    jl_symbol_type->size = 0;
+    jl_symbol_type->abstract = 0;
+    jl_symbol_type->mutabl = 1;
+    jl_symbol_type->ninitialized = 0;
+    jl_precompute_memoized_dt(jl_symbol_type);
 
     jl_simplevector_type->name = jl_new_typename_in(jl_symbol("SimpleVector"), core);
     jl_simplevector_type->name->wrapper = (jl_value_t*)jl_simplevector_type;
-    jl_simplevector_type->name->mt = jl_new_method_table(jl_simplevector_type->name->name, core);
+    jl_simplevector_type->name->mt = jl_nonfunction_mt;
     jl_simplevector_type->super = jl_any_type;
     jl_simplevector_type->parameters = jl_emptysvec;
     jl_simplevector_type->name->names = jl_emptysvec;
@@ -1809,7 +1845,7 @@ void jl_init_types(void) JL_GC_DISABLED
 
     jl_tvar_type = jl_new_datatype(jl_symbol("TypeVar"), core, jl_any_type, jl_emptysvec,
                                    jl_perm_symsvec(3, "name", "lb", "ub"),
-                                   jl_svec(3, jl_sym_type, jl_any_type, jl_any_type),
+                                   jl_svec(3, jl_symbol_type, jl_any_type, jl_any_type),
                                    0, 1, 3);
 
     jl_unionall_type = jl_new_datatype(jl_symbol("UnionAll"), core, type_type, jl_emptysvec,
@@ -1930,8 +1966,8 @@ void jl_init_types(void) JL_GC_DISABLED
                             jl_type_type, // TupleType
                             jl_any_type, // TupleType
                             jl_any_type, // SimpleVector{TupleType}
-                            jl_long_type, // Int
-                            jl_long_type, // Int
+                            jl_ulong_type, // UInt
+                            jl_ulong_type, // UInt
                             jl_any_type, // Any
                             jl_bool_type,
                             jl_bool_type,
@@ -1940,6 +1976,8 @@ void jl_init_types(void) JL_GC_DISABLED
 
     jl_function_type = jl_new_abstracttype((jl_value_t*)jl_symbol("Function"), core, jl_any_type, jl_emptysvec);
     jl_builtin_type  = jl_new_abstracttype((jl_value_t*)jl_symbol("Builtin"), core, jl_function_type, jl_emptysvec);
+    jl_function_type->name->mt = NULL; // subtypes of Function have independent method tables
+    jl_builtin_type->name->mt = NULL;  // so they don't share the Any type table
 
     tv = jl_svec2(tvar("T"), tvar("N"));
     jl_abstractarray_type = (jl_unionall_t*)
@@ -1963,7 +2001,7 @@ void jl_init_types(void) JL_GC_DISABLED
     jl_compute_field_offsets((jl_datatype_t*)jl_unwrap_unionall((jl_value_t*)jl_array_type));
 
     jl_array_any_type = jl_apply_type2((jl_value_t*)jl_array_type, (jl_value_t*)jl_any_type, jl_box_long(1));
-    jl_array_symbol_type = jl_apply_type2((jl_value_t*)jl_array_type, (jl_value_t*)jl_sym_type, jl_box_long(1));
+    jl_array_symbol_type = jl_apply_type2((jl_value_t*)jl_array_type, (jl_value_t*)jl_symbol_type, jl_box_long(1));
     jl_array_uint8_type = jl_apply_type2((jl_value_t*)jl_array_type, (jl_value_t*)jl_uint8_type, jl_box_long(1));
     jl_array_int32_type = jl_apply_type2((jl_value_t*)jl_array_type, (jl_value_t*)jl_int32_type, jl_box_long(1));
 
@@ -1971,13 +2009,13 @@ void jl_init_types(void) JL_GC_DISABLED
         jl_new_datatype(jl_symbol("Expr"), core,
                         jl_any_type, jl_emptysvec,
                         jl_perm_symsvec(2, "head", "args"),
-                        jl_svec(2, jl_sym_type, jl_array_any_type),
+                        jl_svec(2, jl_symbol_type, jl_array_any_type),
                         0, 1, 2);
 
     jl_module_type =
         jl_new_datatype(jl_symbol("Module"), core, jl_any_type, jl_emptysvec,
                         jl_perm_symsvec(2, "name", "parent"),
-                        jl_svec(2, jl_sym_type, jl_any_type), 0, 1, 2);
+                        jl_svec(2, jl_symbol_type, jl_any_type), 0, 1, 2);
 
     jl_linenumbernode_type =
         jl_new_datatype(jl_symbol("LineNumberNode"), core, jl_any_type, jl_emptysvec,
@@ -1987,7 +2025,7 @@ void jl_init_types(void) JL_GC_DISABLED
     jl_lineinfonode_type =
         jl_new_datatype(jl_symbol("LineInfoNode"), core, jl_any_type, jl_emptysvec,
                         jl_perm_symsvec(4, "method", "file", "line", "inlined_at"),
-                        jl_svec(4, jl_any_type, jl_sym_type, jl_long_type, jl_long_type), 0, 0, 4);
+                        jl_svec(4, jl_any_type, jl_symbol_type, jl_long_type, jl_long_type), 0, 0, 4);
 
     jl_gotonode_type =
         jl_new_datatype(jl_symbol("GotoNode"), core, jl_any_type, jl_emptysvec,
@@ -2027,12 +2065,12 @@ void jl_init_types(void) JL_GC_DISABLED
     jl_globalref_type =
         jl_new_datatype(jl_symbol("GlobalRef"), core, jl_any_type, jl_emptysvec,
                         jl_perm_symsvec(2, "mod", "name"),
-                        jl_svec(2, jl_module_type, jl_sym_type), 0, 0, 2);
+                        jl_svec(2, jl_module_type, jl_symbol_type), 0, 0, 2);
 
     jl_code_info_type =
         jl_new_datatype(jl_symbol("CodeInfo"), core,
                         jl_any_type, jl_emptysvec,
-                        jl_perm_symsvec(17,
+                        jl_perm_symsvec(18,
                             "code",
                             "codelocs",
                             "ssavaluetypes",
@@ -2044,13 +2082,14 @@ void jl_init_types(void) JL_GC_DISABLED
                             "slottypes",
                             "rettype",
                             "parent",
+                            "edges",
                             "min_world",
                             "max_world",
                             "inferred",
                             "inlineable",
                             "propagate_inbounds",
                             "pure"),
-                        jl_svec(17,
+                        jl_svec(18,
                             jl_array_any_type,
                             jl_any_type,
                             jl_any_type,
@@ -2062,18 +2101,19 @@ void jl_init_types(void) JL_GC_DISABLED
                             jl_any_type,
                             jl_any_type,
                             jl_any_type,
-                            jl_long_type,
-                            jl_long_type,
+                            jl_any_type,
+                            jl_ulong_type,
+                            jl_ulong_type,
                             jl_bool_type,
                             jl_bool_type,
                             jl_bool_type,
                             jl_bool_type),
-                        0, 1, 17);
+                        0, 1, 18);
 
     jl_method_type =
         jl_new_datatype(jl_symbol("Method"), core,
                         jl_any_type, jl_emptysvec,
-                        jl_perm_symsvec(20,
+                        jl_perm_symsvec(21,
                             "name",
                             "module",
                             "file",
@@ -2082,6 +2122,7 @@ void jl_init_types(void) JL_GC_DISABLED
                             "deleted_world",
                             "sig",
                             "ambig",
+                            "resorted",
                             "specializations",
                             "slot_syms",
                             "source",
@@ -2094,15 +2135,16 @@ void jl_init_types(void) JL_GC_DISABLED
                             "nospecialize",
                             "isva",
                             "pure"),
-                        jl_svec(20,
-                            jl_sym_type,
+                        jl_svec(21,
+                            jl_symbol_type,
                             jl_module_type,
-                            jl_sym_type,
+                            jl_symbol_type,
                             jl_int32_type,
-                            jl_long_type,
-                            jl_long_type,
+                            jl_ulong_type,
+                            jl_ulong_type,
                             jl_type_type,
-                            jl_any_type, // Union{Array, Nothing}
+                            jl_any_type, // Union{Vector, Nothing}
+                            jl_any_type, // Union{Vector, Nothing}
                             jl_any_type, // TypeMap
                             jl_string_type,
                             jl_any_type,
@@ -2115,7 +2157,7 @@ void jl_init_types(void) JL_GC_DISABLED
                             jl_int32_type,
                             jl_bool_type,
                             jl_bool_type),
-                        0, 1, 10);
+                        0, 1, 11);
 
     jl_method_instance_type =
         jl_new_datatype(jl_symbol("MethodInstance"), core,
@@ -2156,8 +2198,8 @@ void jl_init_types(void) JL_GC_DISABLED
                         jl_svec(11,
                             jl_method_instance_type,
                             jl_any_type,
-                            jl_long_type,
-                            jl_long_type,
+                            jl_ulong_type,
+                            jl_ulong_type,
                             jl_any_type,
                             jl_any_type,
                             jl_any_type,
@@ -2168,9 +2210,9 @@ void jl_init_types(void) JL_GC_DISABLED
                         0, 1, 1);
     jl_svecset(jl_code_instance_type->types, 1, jl_code_instance_type);
 
-    // all kinds of types share a method table
+    // all Kinds share the Type method table (not the nonfunction one)
     jl_unionall_type->name->mt = jl_uniontype_type->name->mt = jl_datatype_type->name->mt =
-        jl_type_typename->mt;
+        jl_type_type_mt;
 
     jl_intrinsic_type = jl_new_primitivetype((jl_value_t*)jl_symbol("IntrinsicFunction"), core,
                                              jl_builtin_type, jl_emptysvec, 32);
@@ -2208,6 +2250,39 @@ void jl_init_types(void) JL_GC_DISABLED
     ((jl_datatype_t*)jl_unwrap_unionall((jl_value_t*)jl_namedtuple_type))->layout = NULL;
     jl_namedtuple_typename = ntt->name;
 
+    jl_task_type = (jl_datatype_t*)
+        jl_new_datatype(jl_symbol("Task"),
+                        NULL,
+                        jl_any_type,
+                        jl_emptysvec,
+                        jl_perm_symsvec(11,
+                                        "next",
+                                        "queue",
+                                        "storage",
+                                        "state",
+                                        "donenotify",
+                                        "result",
+                                        "exception",
+                                        "backtrace",
+                                        "logstate",
+                                        "code",
+                                        "sticky"),
+                        jl_svec(11,
+                                jl_any_type,
+                                jl_any_type,
+                                jl_any_type,
+                                jl_symbol_type,
+                                jl_any_type,
+                                jl_any_type,
+                                jl_any_type,
+                                jl_any_type,
+                                jl_any_type,
+                                jl_any_type,
+                                jl_bool_type),
+                        0, 1, 9);
+    jl_value_t *listt = jl_new_struct(jl_uniontype_type, jl_task_type, jl_void_type);
+    jl_svecset(jl_task_type->types, 0, listt);
+
     // complete builtin type metadata
     jl_value_t *pointer_void = jl_apply_type1((jl_value_t*)jl_pointer_type, (jl_value_t*)jl_void_type);
     jl_voidpointer_type = (jl_datatype_t*)pointer_void;
@@ -2223,8 +2298,9 @@ void jl_init_types(void) JL_GC_DISABLED
     jl_svecset(jl_datatype_type->types, 15, jl_bool_type);
     jl_svecset(jl_datatype_type->types, 16, jl_bool_type);
     jl_svecset(jl_datatype_type->types, 17, jl_bool_type);
-    jl_svecset(jl_datatype_type->types, 18, jl_voidpointer_type);
+    jl_svecset(jl_datatype_type->types, 18, jl_bool_type);
     jl_svecset(jl_datatype_type->types, 19, jl_voidpointer_type);
+    jl_svecset(jl_datatype_type->types, 20, jl_voidpointer_type);
     jl_svecset(jl_typename_type->types, 1, jl_module_type);
     jl_svecset(jl_typename_type->types, 6, jl_long_type);
     jl_svecset(jl_typename_type->types, 3, jl_type_type);
@@ -2239,7 +2315,8 @@ void jl_init_types(void) JL_GC_DISABLED
     jl_svecset(jl_methtable_type->types, 8, jl_int32_type); // uint32_t
 #endif
     jl_svecset(jl_methtable_type->types, 9, jl_uint8_type);
-    jl_svecset(jl_method_type->types, 11, jl_method_instance_type);
+    jl_svecset(jl_methtable_type->types, 10, jl_uint8_type);
+    jl_svecset(jl_method_type->types, 12, jl_method_instance_type);
     jl_svecset(jl_method_instance_type->types, 5, jl_code_instance_type);
     jl_svecset(jl_code_instance_type->types, 7, jl_voidpointer_type);
     jl_svecset(jl_code_instance_type->types, 8, jl_voidpointer_type);
@@ -2263,7 +2340,7 @@ void jl_init_types(void) JL_GC_DISABLED
     jl_compute_field_offsets(jl_code_instance_type);
     jl_compute_field_offsets(jl_unionall_type);
     jl_compute_field_offsets(jl_simplevector_type);
-    jl_compute_field_offsets(jl_sym_type);
+    jl_compute_field_offsets(jl_symbol_type);
 }
 
 #ifdef __cplusplus
