@@ -215,10 +215,8 @@ JL_DLLEXPORT void jl_enter_handler(jl_handler_t *eh)
     // Must have no safepoint
     eh->prev = current_task->eh;
     eh->gcstack = ptls->pgcstack;
-#ifdef JULIA_ENABLE_THREADING
     eh->gc_state = ptls->gc_state;
     eh->locks_len = current_task->locks.len;
-#endif
     eh->defer_signal = ptls->defer_signal;
     eh->finalizers_inhibited = ptls->finalizers_inhibited;
     eh->world_age = ptls->world_age;
@@ -250,14 +248,12 @@ JL_DLLEXPORT void jl_eh_restore_state(jl_handler_t *eh)
     int8_t old_gc_state = ptls->gc_state;
     current_task->eh = eh->prev;
     ptls->pgcstack = eh->gcstack;
-#ifdef JULIA_ENABLE_THREADING
     arraylist_t *locks = &current_task->locks;
     if (locks->len > eh->locks_len) {
         for (size_t i = locks->len;i > eh->locks_len;i--)
             jl_mutex_unlock_nogc((jl_mutex_t*)locks->items[i - 1]);
         locks->len = eh->locks_len;
     }
-#endif
     ptls->world_age = eh->world_age;
     ptls->defer_signal = eh->defer_signal;
     ptls->gc_state = eh->gc_state;
@@ -332,29 +328,6 @@ void jl_push_excstack(jl_excstack_t **stack JL_REQUIRE_ROOTED_SLOT JL_ROOTING_AR
     jl_excstack_raw(s)[s->top-1] = (uintptr_t)exception;
 }
 
-// misc -----------------------------------------------------------------------
-
-// perform f(args...) on stack
-JL_DLLEXPORT jl_value_t *jl_apply_2va(jl_value_t *f, jl_value_t **args, uint32_t nargs)
-{
-    nargs++;
-    int onstack = (nargs < jl_page_size/sizeof(jl_value_t*));
-    jl_value_t **newargs;
-    JL_GC_PUSHARGS(newargs, onstack ? nargs : 1);
-    jl_svec_t *arg_heap = NULL;
-    newargs[0] = f;  // make sure f is rooted
-    if (!onstack) {
-        arg_heap = jl_alloc_svec(nargs);
-        newargs[0] = (jl_value_t*)arg_heap;
-        newargs = jl_svec_data(arg_heap);
-        newargs[0] = f;
-    }
-    memcpy(&newargs[1], args, (nargs-1)*sizeof(jl_value_t*));
-    jl_value_t *ret = jl_apply_generic(newargs, nargs);
-    JL_GC_POP();
-    return ret;
-}
-
 // conversion -----------------------------------------------------------------
 
 JL_DLLEXPORT void *(jl_symbol_name)(jl_sym_t *s)
@@ -524,14 +497,14 @@ JL_DLLEXPORT void jl_flush_cstdio(void) JL_NOTSAFEPOINT
 JL_DLLEXPORT jl_value_t *jl_stdout_obj(void) JL_NOTSAFEPOINT
 {
     if (jl_base_module == NULL) return NULL;
-    jl_value_t *stdout_obj = jl_get_global(jl_base_module, jl_symbol("stdout"));
+    jl_value_t *stdout_obj = jl_get_module_binding(jl_base_module, jl_symbol("stdout"))->value;
     return stdout_obj;
 }
 
 JL_DLLEXPORT jl_value_t *jl_stderr_obj(void) JL_NOTSAFEPOINT
 {
     if (jl_base_module == NULL) return NULL;
-    jl_value_t *stderr_obj = jl_get_global(jl_base_module, jl_symbol("stderr"));
+    jl_value_t *stderr_obj = jl_get_module_binding(jl_base_module, jl_symbol("stderr"))->value;
     return stderr_obj;
 }
 
@@ -558,6 +531,7 @@ struct recur_list {
 };
 
 static size_t jl_static_show_x(JL_STREAM *out, jl_value_t *v, struct recur_list *depth) JL_NOTSAFEPOINT;
+static size_t jl_static_show_next_(JL_STREAM *out, jl_value_t *v, jl_value_t *prev, struct recur_list *depth) JL_NOTSAFEPOINT;
 
 JL_DLLEXPORT int jl_id_start_char(uint32_t wc) JL_NOTSAFEPOINT;
 JL_DLLEXPORT int jl_id_char(uint32_t wc) JL_NOTSAFEPOINT;
@@ -575,6 +549,51 @@ JL_DLLEXPORT int jl_is_identifier(char *str) JL_NOTSAFEPOINT
     return 1;
 }
 
+static jl_datatype_t *first_arg_datatype(jl_value_t *a JL_PROPAGATES_ROOT, int got_tuple1) JL_NOTSAFEPOINT
+{
+    if (jl_is_datatype(a)) {
+        if (got_tuple1)
+            return (jl_datatype_t*)a;
+        if (jl_is_tuple_type(a)) {
+            if (jl_nparams(a) < 1)
+                return NULL;
+            return first_arg_datatype(jl_tparam0(a), 1);
+        }
+        return NULL;
+    }
+    else if (jl_is_typevar(a)) {
+        return first_arg_datatype(((jl_tvar_t*)a)->ub, got_tuple1);
+    }
+    else if (jl_is_unionall(a)) {
+        return first_arg_datatype(((jl_unionall_t*)a)->body, got_tuple1);
+    }
+    else if (jl_is_uniontype(a)) {
+        jl_uniontype_t *u = (jl_uniontype_t*)a;
+        jl_datatype_t *d1 = first_arg_datatype(u->a, got_tuple1);
+        if (d1 == NULL) return NULL;
+        jl_datatype_t *d2 = first_arg_datatype(u->b, got_tuple1);
+        if (d2 == NULL || d1->name != d2->name)
+            return NULL;
+        return d1;
+    }
+    return NULL;
+}
+
+// get DataType of first tuple element (if present), or NULL if cannot be determined
+JL_DLLEXPORT jl_datatype_t *jl_first_argument_datatype(jl_value_t *argtypes JL_PROPAGATES_ROOT) JL_NOTSAFEPOINT
+{
+    return first_arg_datatype(argtypes, 0);
+}
+
+// get DataType implied by a single given type, or `nothing`
+JL_DLLEXPORT jl_value_t *jl_argument_datatype(jl_value_t *argt JL_PROPAGATES_ROOT) JL_NOTSAFEPOINT
+{
+    jl_datatype_t *dt = first_arg_datatype(argt, 1);
+    if (dt == NULL)
+        return jl_nothing;
+    return (jl_value_t*)dt;
+}
+
 // `v` might be pointing to a field inlined in a structure therefore
 // `jl_typeof(v)` may not be the same with `vt` and only `vt` should be
 // used to determine the type of the value.
@@ -600,7 +619,7 @@ static size_t jl_static_show_x_(JL_STREAM *out, jl_value_t *v, jl_datatype_t *vt
     else if (v == (jl_value_t*)jl_typename_type) {
         n += jl_printf(out, "Core.TypeName");
     }
-    else if (v == (jl_value_t*)jl_sym_type) {
+    else if (v == (jl_value_t*)jl_symbol_type) {
         n += jl_printf(out, "Symbol");
     }
     else if (v == (jl_value_t*)jl_methtable_type) {
@@ -650,7 +669,7 @@ static size_t jl_static_show_x_(JL_STREAM *out, jl_value_t *v, jl_datatype_t *vt
             !strchr(jl_symbol_name(globname), '@') && dv->name->module &&
             jl_binding_resolved_p(dv->name->module, globname)) {
             jl_binding_t *b = jl_get_module_binding(dv->name->module, globname);
-            if (b && jl_typeof(b->value) == v)
+            if (b && b->value && jl_typeof(b->value) == v)
                 globfunc = 1;
         }
         jl_sym_t *sym = globfunc ? globname : dv->name->name;
@@ -820,7 +839,7 @@ static size_t jl_static_show_x_(JL_STREAM *out, jl_value_t *v, jl_datatype_t *vt
         }
         n += jl_printf(out, "%s", jl_symbol_name(m->name));
     }
-    else if (vt == jl_sym_type) {
+    else if (vt == jl_symbol_type) {
         char *sn = jl_symbol_name((jl_sym_t*)v);
         int quoted = !jl_is_identifier(sn) && jl_operator_precedence(sn) == 0;
         if (quoted)
@@ -979,7 +998,7 @@ static size_t jl_static_show_x_(JL_STREAM *out, jl_value_t *v, jl_datatype_t *vt
                     n += jl_static_show_x(out, *(jl_value_t**)fld_ptr, depth);
                 }
                 else {
-                    jl_datatype_t *ft = (jl_datatype_t*)jl_field_type(vt, i);
+                    jl_datatype_t *ft = (jl_datatype_t*)jl_field_type_concrete(vt, i);
                     if (jl_is_uniontype(ft)) {
                         uint8_t sel = ((uint8_t*)fld_ptr)[jl_field_size(vt, i) - 1];
                         ft = (jl_datatype_t*)jl_nth_union_component((jl_value_t*)ft, sel);
@@ -993,7 +1012,7 @@ static size_t jl_static_show_x_(JL_STREAM *out, jl_value_t *v, jl_datatype_t *vt
             }
             if (vt == jl_typemap_entry_type) {
                 n += jl_printf(out, ", next=↩︎\n  ");
-                n += jl_static_show_x(out, jl_fieldref_noalloc(v, 0), depth);
+                n += jl_static_show_next_(out, (jl_value_t*)((jl_typemap_entry_t*)v)->next, v, depth);
             }
         }
         n += jl_printf(out, ")");
@@ -1009,6 +1028,13 @@ static size_t jl_static_show_x_(JL_STREAM *out, jl_value_t *v, jl_datatype_t *vt
 static size_t jl_static_show_x(JL_STREAM *out, jl_value_t *v, struct recur_list *depth) JL_NOTSAFEPOINT
 {
     // show values without calling a julia method or allocating through the GC
+    return jl_static_show_next_(out, v, NULL, depth);
+}
+
+static size_t jl_static_show_next_(JL_STREAM *out, jl_value_t *v, jl_value_t *prev, struct recur_list *depth) JL_NOTSAFEPOINT
+{
+    // helper for showing a typemap list by following the next pointers
+    // while being careful about avoiding any recursion due to malformed (circular) references
     if (v == NULL) {
         return jl_printf(out, "#<null>");
     }
@@ -1016,14 +1042,48 @@ static size_t jl_static_show_x(JL_STREAM *out, jl_value_t *v, struct recur_list
         return jl_printf(out, "#<%d>", (int)(uintptr_t)v);
     }
     unsigned int dist = 1;
-    struct recur_list this_item = {depth, v}, *p = depth;
+    struct recur_list this_item = {depth, v},
+                      *newdepth = &this_item,
+                      *p = depth;
     while (p) {
+        if (jl_typeis(v, jl_typemap_entry_type) && newdepth == &this_item) {
+            jl_value_t *m = p->v;
+            unsigned nid = 1;
+            while (m && jl_typeis(m, jl_typemap_entry_type)) {
+                if (m == v) {
+                    return jl_printf(out, "<typemap reference #%u @-%u ", nid, dist) +
+                           jl_static_show_x(out, (jl_value_t*)((jl_typemap_entry_t*)m)->sig, depth) +
+                           jl_printf(out, ">");
+                }
+                if (m == prev) {
+                    newdepth = depth;
+                    break;
+                }
+                // verify that we aren't trying to follow a circular list
+                // by following the list again, and ensuring this is the only link to next
+                jl_value_t *mnext = (jl_value_t*)((jl_typemap_entry_t*)m)->next;
+                jl_value_t *m2 = p->v;
+                if (m2 == mnext)
+                    break;
+                while (m2 && jl_typeis(m2, jl_typemap_entry_type)) {
+                    jl_value_t *mnext2 = (jl_value_t*)((jl_typemap_entry_t*)m2)->next;
+                    if (mnext2 == mnext) {
+                        if (m2 != m)
+                            mnext = NULL;
+                        break;
+                    }
+                    m2 = mnext2;
+                }
+                m = mnext;
+                nid++;
+            }
+        }
         if (p->v == v)
             return jl_printf(out, "<circular reference @-%u>", dist);
         dist++;
         p = p->prev;
     }
-    return jl_static_show_x_(out, v, (jl_datatype_t*)jl_typeof(v), &this_item);
+    return jl_static_show_x_(out, v, (jl_datatype_t*)jl_typeof(v), newdepth);
 }
 
 JL_DLLEXPORT size_t jl_static_show(JL_STREAM *out, jl_value_t *v) JL_NOTSAFEPOINT
@@ -1088,7 +1148,7 @@ JL_DLLEXPORT size_t jl_static_show_func_sig(JL_STREAM *s, jl_value_t *type) JL_N
     return n;
 }
 
-JL_DLLEXPORT void jl_(void *jl_value)
+JL_DLLEXPORT void jl_(void *jl_value) JL_NOTSAFEPOINT
 {
     jl_ptls_t ptls = jl_get_ptls_states();
     jl_jmp_buf *old_buf = ptls->safe_restore;
