@@ -18,7 +18,7 @@ const unsigned int host_char_bit = 8;
 JL_DLLEXPORT jl_value_t *jl_bitcast(jl_value_t *ty, jl_value_t *v)
 {
     JL_TYPECHK(bitcast, datatype, ty);
-    if (!jl_is_leaf_type(ty) || !jl_is_primitivetype(ty))
+    if (!jl_is_concrete_type(ty) || !jl_is_primitivetype(ty))
         jl_error("bitcast: target type not a leaf primitive type");
     if (!jl_is_primitivetype(jl_typeof(v)))
         jl_error("bitcast: value not a primitive type");
@@ -37,7 +37,6 @@ JL_DLLEXPORT jl_value_t *jl_pointerref(jl_value_t *p, jl_value_t *i, jl_value_t
     JL_TYPECHK(pointerref, pointer, p);
     JL_TYPECHK(pointerref, long, i)
     JL_TYPECHK(pointerref, long, align);
-    // TODO: alignment
     jl_value_t *ety = jl_tparam0(jl_typeof(p));
     if (ety == (jl_value_t*)jl_any_type) {
         jl_value_t **pp = (jl_value_t**)(jl_unbox_long(p) + (jl_unbox_long(i)-1)*sizeof(void*));
@@ -58,7 +57,6 @@ JL_DLLEXPORT jl_value_t *jl_pointerset(jl_value_t *p, jl_value_t *x, jl_value_t
     JL_TYPECHK(pointerset, pointer, p);
     JL_TYPECHK(pointerset, long, i);
     JL_TYPECHK(pointerref, long, align);
-    // TODO: alignment
     jl_value_t *ety = jl_tparam0(jl_typeof(p));
     if (ety == (jl_value_t*)jl_any_type) {
         jl_value_t **pp = (jl_value_t**)(jl_unbox_long(p) + (jl_unbox_long(i)-1)*sizeof(void*));
@@ -67,11 +65,12 @@ JL_DLLEXPORT jl_value_t *jl_pointerset(jl_value_t *p, jl_value_t *x, jl_value_t
     else {
         if (!jl_is_datatype(ety))
             jl_error("pointerset: invalid pointer");
-        size_t nb = LLT_ALIGN(jl_datatype_size(ety), jl_datatype_align(ety));
+        size_t elsz = jl_datatype_size(ety);
+        size_t nb = LLT_ALIGN(elsz, jl_datatype_align(ety));
         char *pp = (char*)jl_unbox_long(p) + (jl_unbox_long(i)-1)*nb;
         if (jl_typeof(x) != ety)
-            jl_error("pointerset: type mismatch in assign");
-        jl_assign_bits(pp, x);
+            jl_type_error("pointerset", ety, x);
+        memcpy(pp, x, elsz);
     }
     return p;
 }
@@ -79,22 +78,27 @@ JL_DLLEXPORT jl_value_t *jl_pointerset(jl_value_t *p, jl_value_t *x, jl_value_t
 JL_DLLEXPORT jl_value_t *jl_cglobal(jl_value_t *v, jl_value_t *ty)
 {
     JL_TYPECHK(cglobal, type, ty);
+    JL_GC_PUSH1(&v);
     jl_value_t *rt =
-        v == (jl_value_t*)jl_void_type ? (jl_value_t*)jl_voidpointer_type : // a common case
+        ty == (jl_value_t*)jl_void_type ? (jl_value_t*)jl_voidpointer_type : // a common case
             (jl_value_t*)jl_apply_type1((jl_value_t*)jl_pointer_type, ty);
+    JL_GC_PROMISE_ROOTED(rt); // (JL_ALWAYS_LEAFTYPE)
 
-    if (!jl_is_leaf_type(rt))
-        jl_error("cglobal: type argument not a leaftype");
+    if (!jl_is_concrete_type(rt))
+        jl_error("cglobal: type argument not concrete");
 
     if (jl_is_tuple(v) && jl_nfields(v) == 1)
         v = jl_fieldref(v, 0);
 
-    if (jl_is_pointer(v))
-        return jl_bitcast(rt, v);
+    if (jl_is_pointer(v)) {
+        v = jl_bitcast(rt, v);
+        JL_GC_POP();
+        return v;
+    }
 
     char *f_lib = NULL;
     if (jl_is_tuple(v) && jl_nfields(v) > 1) {
-        jl_value_t *t1 = jl_fieldref(v, 1);
+        jl_value_t *t1 = jl_fieldref_noalloc(v, 1);
         v = jl_fieldref(v, 0);
         if (jl_is_symbol(t1))
             f_lib = jl_symbol_name((jl_sym_t*)t1);
@@ -117,10 +121,12 @@ JL_DLLEXPORT jl_value_t *jl_cglobal(jl_value_t *v, jl_value_t *ty)
         f_lib = (char*)jl_dlfind_win32(f_name);
 #endif
 
-    void *ptr = jl_dlsym(jl_get_library(f_lib), f_name);
+    void *ptr;
+    jl_dlsym(jl_get_library(f_lib), f_name, &ptr, 1);
     jl_value_t *jv = jl_gc_alloc_1w();
     jl_set_typeof(jv, rt);
     *(void**)jl_data_ptr(jv) = ptr;
+    JL_GC_POP();
     return jv;
 }
 
@@ -128,19 +134,19 @@ JL_DLLEXPORT jl_value_t *jl_cglobal_auto(jl_value_t *v) {
     return jl_cglobal(v, (jl_value_t*)jl_void_type);
 }
 
-static inline char signbitbyte(void *a, unsigned bytes)
+static inline char signbitbyte(void *a, unsigned bytes) JL_NOTSAFEPOINT
 {
     // sign bit of an signed number of n bytes, as a byte
     return (((signed char*)a)[bytes - 1] < 0) ? ~0 : 0;
 }
 
-static inline char usignbitbyte(void *a, unsigned bytes)
+static inline char usignbitbyte(void *a, unsigned bytes) JL_NOTSAFEPOINT
 {
     // sign bit of an unsigned number
     return 0;
 }
 
-static inline unsigned select_by_size(unsigned sz)
+static inline unsigned select_by_size(unsigned sz) JL_NOTSAFEPOINT
 {
     /* choose the right sized function specialization */
     switch (sz) {
@@ -155,7 +161,7 @@ static inline unsigned select_by_size(unsigned sz)
 
 #define SELECTOR_FUNC(intrinsic) \
     typedef intrinsic##_t select_##intrinsic##_t[6]; \
-    static inline intrinsic##_t select_##intrinsic(unsigned sz, const select_##intrinsic##_t list) \
+    static inline intrinsic##_t select_##intrinsic(unsigned sz, const select_##intrinsic##_t list) JL_NOTSAFEPOINT \
     { \
         intrinsic##_t thunk = list[select_by_size(sz)]; \
         if (!thunk) thunk = list[0]; \
@@ -175,7 +181,7 @@ static inline unsigned select_by_size(unsigned sz)
 // nbits::number of bits
 // c_type::c_type corresponding to nbits
 #define un_iintrinsic_ctype(OP, name, nbits, c_type) \
-static inline void jl_##name##nbits(unsigned runtime_nbits, void *pa, void *pr) \
+static inline void jl_##name##nbits(unsigned runtime_nbits, void *pa, void *pr) JL_NOTSAFEPOINT \
 { \
     c_type a = *(c_type*)pa; \
     *(c_type*)pr = OP(a); \
@@ -187,7 +193,7 @@ static inline void jl_##name##nbits(unsigned runtime_nbits, void *pa, void *pr)
 // nbits::number of bits
 // c_type::c_type corresponding to nbits
 #define uu_iintrinsic_ctype(OP, name, nbits, c_type) \
-static inline unsigned jl_##name##nbits(unsigned runtime_nbits, void *pa) \
+static inline unsigned jl_##name##nbits(unsigned runtime_nbits, void *pa) JL_NOTSAFEPOINT \
 { \
     c_type a = *(c_type*)pa; \
     return OP(a); \
@@ -199,7 +205,7 @@ static inline unsigned jl_##name##nbits(unsigned runtime_nbits, void *pa) \
 // nbits::number of bits in the *input*
 // c_type::c_type corresponding to nbits
 #define un_fintrinsic_ctype(OP, name, c_type) \
-static inline void name(unsigned osize, void *pa, void *pr) \
+static inline void name(unsigned osize, void *pa, void *pr) JL_NOTSAFEPOINT \
 { \
     c_type a = *(c_type*)pa; \
     OP((c_type*)pr, a); \
@@ -211,7 +217,7 @@ static inline void name(unsigned osize, void *pa, void *pr) \
 // nbits::number of bits
 // c_type::c_type corresponding to nbits
 #define bi_intrinsic_ctype(OP, name, nbits, c_type) \
-static void jl_##name##nbits(unsigned runtime_nbits, void *pa, void *pb, void *pr) \
+static void jl_##name##nbits(unsigned runtime_nbits, void *pa, void *pb, void *pr) JL_NOTSAFEPOINT \
 { \
     c_type a = *(c_type*)pa; \
     c_type b = *(c_type*)pb; \
@@ -224,7 +230,7 @@ static void jl_##name##nbits(unsigned runtime_nbits, void *pa, void *pb, void *p
 // nbits::number of bits
 // c_type::c_type corresponding to nbits
 #define bool_intrinsic_ctype(OP, name, nbits, c_type) \
-static int jl_##name##nbits(unsigned runtime_nbits, void *pa, void *pb) \
+static int jl_##name##nbits(unsigned runtime_nbits, void *pa, void *pb) JL_NOTSAFEPOINT \
 { \
     c_type a = *(c_type*)pa; \
     c_type b = *(c_type*)pb; \
@@ -237,14 +243,12 @@ static int jl_##name##nbits(unsigned runtime_nbits, void *pa, void *pb) \
 // nbits::number of bits
 // c_type::c_type corresponding to nbits
 #define checked_intrinsic_ctype(CHECK_OP, OP, name, nbits, c_type) \
-static int jl_##name##nbits(unsigned runtime_nbits, void *pa, void *pb, void *pr) \
+static int jl_##name##nbits(unsigned runtime_nbits, void *pa, void *pb, void *pr) JL_NOTSAFEPOINT \
 { \
     c_type a = *(c_type*)pa; \
     c_type b = *(c_type*)pb; \
-    if (CHECK_OP(a, b)) \
-        return 1; \
     *(c_type*)pr = (c_type)OP(a, b); \
-    return 0; \
+    return CHECK_OP(a, b); \
 }
 
 // float inputs
@@ -253,7 +257,7 @@ static int jl_##name##nbits(unsigned runtime_nbits, void *pa, void *pb, void *pr
 // nbits::number of bits
 // c_type::c_type corresponding to nbits
 #define ter_intrinsic_ctype(OP, name, nbits, c_type) \
-static void jl_##name##nbits(unsigned runtime_nbits, void *pa, void *pb, void *pc, void *pr) \
+static void jl_##name##nbits(unsigned runtime_nbits, void *pa, void *pb, void *pc, void *pr) JL_NOTSAFEPOINT \
 { \
     c_type a = *(c_type*)pa; \
     c_type b = *(c_type*)pb; \
@@ -350,27 +354,26 @@ jl_value_t *jl_iintrinsic_1(jl_value_t *ty, jl_value_t *a, const char *name,
 
 static inline jl_value_t *jl_intrinsiclambda_ty1(jl_value_t *ty, void *pa, unsigned osize, unsigned osize2, const void *voidlist)
 {
-    jl_ptls_t ptls = jl_get_ptls_states();
-    jl_value_t *newv = jl_gc_alloc(ptls, jl_datatype_size(ty), ty);
     intrinsic_1_t op = select_intrinsic_1(osize2, (const intrinsic_1_t*)voidlist);
-    op(osize * host_char_bit, pa, jl_data_ptr(newv));
-    return newv;
+    void *pr = alloca(osize2);
+    op(osize * host_char_bit, pa, pr);
+    return jl_new_bits(ty, pr);
 }
 
 static inline jl_value_t *jl_intrinsiclambda_u1(jl_value_t *ty, void *pa, unsigned osize, unsigned osize2, const void *voidlist)
 {
     jl_ptls_t ptls = jl_get_ptls_states();
-    jl_value_t *newv = jl_gc_alloc(ptls, jl_datatype_size(ty), ty);
     intrinsic_u1_t op = select_intrinsic_u1(osize2, (const intrinsic_u1_t*)voidlist);
-    unsigned cnt = op(osize * host_char_bit, pa);
-    // TODO: the following memset/memcpy assumes little-endian
+    uint64_t cnt = op(osize * host_char_bit, pa);
+    // TODO: the following assume little-endian
     // for big-endian, need to copy from the other end of cnt
-    if (osize > sizeof(unsigned)) {
-        // perform zext, if needed
-        memset((char*)jl_data_ptr(newv) + sizeof(unsigned), 0, osize - sizeof(unsigned));
-        osize = sizeof(unsigned);
+    if (osize <= sizeof(cnt)) {
+        return jl_new_bits(ty, &cnt);
     }
-    memcpy(jl_data_ptr(newv), &cnt, osize);
+    jl_value_t *newv = jl_gc_alloc(ptls, osize, ty);
+    // perform zext, if needed
+    memset((char*)jl_data_ptr(newv) + sizeof(cnt), 0, osize - sizeof(cnt));
+    memcpy(jl_data_ptr(newv), &cnt, sizeof(cnt));
     return newv;
 }
 
@@ -378,17 +381,14 @@ static inline jl_value_t *jl_intrinsiclambda_u1(jl_value_t *ty, void *pa, unsign
 
 typedef void (*intrinsic_cvt_t)(unsigned, void*, unsigned, void*);
 typedef unsigned (*intrinsic_cvt_check_t)(unsigned, unsigned, void*);
-#define cvt_iintrinsic_checked(LLVMOP, check_op, name) \
+#define cvt_iintrinsic(LLVMOP, name) \
 JL_DLLEXPORT jl_value_t *jl_##name(jl_value_t *ty, jl_value_t *a) \
 { \
-    return jl_intrinsic_cvt(ty, a, #name, LLVMOP, check_op); \
+    return jl_intrinsic_cvt(ty, a, #name, LLVMOP); \
 }
-#define cvt_iintrinsic(LLVMOP, name) \
-    cvt_iintrinsic_checked(LLVMOP, NULL, name) \
 
-static inline jl_value_t *jl_intrinsic_cvt(jl_value_t *ty, jl_value_t *a, const char *name, intrinsic_cvt_t op, intrinsic_cvt_check_t check_op)
+static inline jl_value_t *jl_intrinsic_cvt(jl_value_t *ty, jl_value_t *a, const char *name, intrinsic_cvt_t op)
 {
-    jl_ptls_t ptls = jl_get_ptls_states();
     jl_value_t *aty = jl_typeof(a);
     if (!jl_is_primitivetype(aty))
         jl_errorf("%s: value is not a primitive type", name);
@@ -397,14 +397,11 @@ static inline jl_value_t *jl_intrinsic_cvt(jl_value_t *ty, jl_value_t *a, const
     void *pa = jl_data_ptr(a);
     unsigned isize = jl_datatype_size(aty);
     unsigned osize = jl_datatype_size(ty);
-    if (check_op && check_op(isize, osize, pa))
-        jl_throw(jl_inexact_exception);
-    jl_value_t *newv = jl_gc_alloc(ptls, jl_datatype_size(ty), ty);
-    op(aty == (jl_value_t*)jl_bool_type ? 1 : isize * host_char_bit, pa,
-            osize * host_char_bit, jl_data_ptr(newv));
-    if (ty == (jl_value_t*)jl_bool_type)
-        return *(uint8_t*)jl_data_ptr(newv) & 1 ? jl_true : jl_false;
-    return newv;
+    void *pr = alloca(osize);
+    unsigned isize_bits = isize * host_char_bit;
+    unsigned osize_bits = osize * host_char_bit;
+    op(isize_bits, pa, osize_bits, pr);
+    return jl_new_bits(ty, pr);
 }
 
 // floating point
@@ -499,7 +496,7 @@ static const select_intrinsic_cmp_t name##_list = { \
 }; \
 cmp_iintrinsic(name, u)
 
-typedef int (*intrinsic_checked_t)(unsigned, void*, void*, void*);
+typedef int (*intrinsic_checked_t)(unsigned, void*, void*, void*) JL_NOTSAFEPOINT;
 SELECTOR_FUNC(intrinsic_checked)
 #define checked_iintrinsic(name, u, lambda_checked) \
 JL_DLLEXPORT jl_value_t *jl_##name(jl_value_t *a, jl_value_t *b) \
@@ -572,13 +569,10 @@ jl_value_t *jl_iintrinsic_2(jl_value_t *a, jl_value_t *b, const char *name,
 
 static inline jl_value_t *jl_intrinsiclambda_2(jl_value_t *ty, void *pa, void *pb, unsigned sz, unsigned sz2, const void *voidlist)
 {
-    jl_ptls_t ptls = jl_get_ptls_states();
-    jl_value_t *newv = jl_gc_alloc(ptls, jl_datatype_size(ty), ty);
+    void *pr = alloca(sz2);
     intrinsic_2_t op = select_intrinsic_2(sz2, (const intrinsic_2_t*)voidlist);
-    op(sz * host_char_bit, pa, pb, jl_data_ptr(newv));
-    if (ty == (jl_value_t*)jl_bool_type)
-        return *(uint8_t*)jl_data_ptr(newv) & 1 ? jl_true : jl_false;
-    return newv;
+    op(sz * host_char_bit, pa, pb, pr);
+    return jl_new_bits(ty, pr);
 }
 
 static inline jl_value_t *jl_intrinsiclambda_cmp(jl_value_t *ty, void *pa, void *pb, unsigned sz, unsigned sz2, const void *voidlist)
@@ -593,7 +587,8 @@ static inline jl_value_t *jl_intrinsiclambda_checked(jl_value_t *ty, void *pa, v
     jl_value_t *params[2];
     params[0] = ty;
     params[1] = (jl_value_t*)jl_bool_type;
-    jl_datatype_t *tuptyp = jl_apply_tuple_type_v(params,2);
+    jl_datatype_t *tuptyp = jl_apply_tuple_type_v(params, 2);
+    JL_GC_PROMISE_ROOTED(tuptyp); // (JL_ALAWYS_LEAFTYPE)
     jl_ptls_t ptls = jl_get_ptls_states();
     jl_value_t *newv = jl_gc_alloc(ptls, ((jl_datatype_t*)tuptyp)->size, tuptyp);
 
@@ -606,16 +601,12 @@ static inline jl_value_t *jl_intrinsiclambda_checked(jl_value_t *ty, void *pa, v
 }
 static inline jl_value_t *jl_intrinsiclambda_checkeddiv(jl_value_t *ty, void *pa, void *pb, unsigned sz, unsigned sz2, const void *voidlist)
 {
-    jl_ptls_t ptls = jl_get_ptls_states();
-    jl_value_t *newv = jl_gc_alloc(ptls, jl_datatype_size(ty), ty);
+    void *pr = alloca(sz2);
     intrinsic_checked_t op = select_intrinsic_checked(sz2, (const intrinsic_checked_t*)voidlist);
-    int ovflw = op(sz * host_char_bit, pa, pb, jl_data_ptr(newv));
+    int ovflw = op(sz * host_char_bit, pa, pb, pr);
     if (ovflw)
         jl_throw(jl_diverror_exception);
-    if (ty == (jl_value_t*)jl_bool_type)
-        return *(uint8_t*)jl_data_ptr(newv) & 1 ? jl_true : jl_false;
-
-    return newv;
+    return jl_new_bits(ty, pr);
 }
 
 // floating point
@@ -709,8 +700,10 @@ JL_DLLEXPORT jl_value_t *jl_##name(jl_value_t *a, jl_value_t *b, jl_value_t *c)
 un_iintrinsic_fast(LLVMNeg, neg, neg_int, u)
 #define add(a,b) a + b
 bi_iintrinsic_fast(LLVMAdd, add, add_int, u)
+bi_iintrinsic_fast(LLVMAdd, add, add_ptr, u)
 #define sub(a,b) a - b
 bi_iintrinsic_fast(LLVMSub, sub, sub_int, u)
+bi_iintrinsic_fast(LLVMSub, sub, sub_ptr, u)
 #define mul(a,b) a * b
 bi_iintrinsic_fast(LLVMMul, mul, mul_int, u)
 #define div(a,b) a / b
@@ -762,7 +755,7 @@ typedef union {
 } bits64;
 
 #define fpiseq_n(c_type, nbits) \
-static inline int fpiseq##nbits(c_type a, c_type b) { \
+static inline int fpiseq##nbits(c_type a, c_type b) JL_NOTSAFEPOINT { \
     bits##nbits ua, ub; \
     ua.f = a; \
     ub.f = b; \
@@ -773,21 +766,22 @@ fpiseq_n(double, 64)
 #define fpiseq(a,b) \
     sizeof(a) == sizeof(float) ? fpiseq32(a, b) : fpiseq64(a, b)
 
-#define fpislt_n(c_type, nbits) \
-static inline int fpislt##nbits(c_type a, c_type b) { \
-    bits##nbits ua, ub; \
-    ua.f = a; \
-    ub.f = b; \
-    if (!isnan(a) && isnan(b)) \
-        return 1; \
-    if (isnan(a) || isnan(b)) \
-        return 0; \
-    if (ua.d >= 0 && ua.d < ub.d) \
-        return 1; \
-    if (ua.d < 0 && ua.ud > ub.ud) \
-        return 1; \
-    return 0; \
-}
+#define fpislt_n(c_type, nbits)                                         \
+    static inline int fpislt##nbits(c_type a, c_type b) JL_NOTSAFEPOINT \
+    {                                                                   \
+        bits##nbits ua, ub;                                             \
+        ua.f = a;                                                       \
+        ub.f = b;                                                       \
+        if (!isnan(a) && isnan(b))                                      \
+            return 1;                                                   \
+        if (isnan(a) || isnan(b))                                       \
+            return 0;                                                   \
+        if (ua.d >= 0 && ua.d < ub.d)                                   \
+            return 1;                                                   \
+        if (ua.d < 0 && ua.ud > ub.ud)                                  \
+            return 1;                                                   \
+        return 0;                                                       \
+    }
 fpislt_n(float, 32)
 fpislt_n(double, 64)
 #define fpislt(a, b) \
@@ -811,7 +805,7 @@ bi_iintrinsic_fast(LLVMXor, xor_op, xor_int, u)
 bi_iintrinsic_cnvtb_fast(LLVMShl, shl_op, shl_int, u, 1)
 #define lshr_op(a,b) (b >= 8 * sizeof(a)) ? 0 : a >> b
 bi_iintrinsic_cnvtb_fast(LLVMLShr, lshr_op, lshr_int, u, 1)
-#define ashr_op(a,b) ((b < 0 || b >= 8 * sizeof(a)) ? a >> (8*sizeof(a) - 1) : a >> b)
+#define ashr_op(a,b) ((b < 0 || b >= 8 * sizeof(a)) ? a >> (8 * sizeof(a) - 1) : a >> b)
 bi_iintrinsic_cnvtb_fast(LLVMAShr, ashr_op, ashr_int, , 1)
 //#define bswap_op(a) __builtin_bswap(a)
 //un_iintrinsic_fast(LLVMByteSwap, bswap_op, bswap_int, u)
@@ -837,61 +831,61 @@ cvt_iintrinsic(LLVMUItoFP, uitofp)
 cvt_iintrinsic(LLVMFPtoSI, fptosi)
 cvt_iintrinsic(LLVMFPtoUI, fptoui)
 
-#define fpcvt(pr, a) \
+#define fptrunc(pr, a) \
+        if (!(osize < 8 * sizeof(a))) \
+            jl_error("fptrunc: output bitsize must be < input bitsize"); \
         if (osize == 32) \
             *(float*)pr = a; \
         else if (osize == 64) \
             *(double*)pr = a; \
         else \
-            jl_error("fptrunc/fpext: runtime floating point intrinsics are not implemented for bit sizes other than 32 and 64");
-un_fintrinsic_withtype(fpcvt,fptrunc)
-un_fintrinsic_withtype(fpcvt,fpext)
-
-// checked conversion
-static inline int all_eq(char *p, char n, char v)
-{
-    // computes p[0:n] == v
-    while (n--)
-        if (*p++ != v)
-            return 0;
-    return 1;
-}
-static unsigned check_trunc_sint(unsigned isize, unsigned osize, void *pa)
-{
-    return !all_eq((char*)pa + osize, isize - osize, signbitbyte(pa, isize)); // TODO: assumes little-endian
-}
-cvt_iintrinsic_checked(LLVMTrunc, check_trunc_sint, checked_trunc_sint)
-static unsigned check_trunc_uint(unsigned isize, unsigned osize, void *pa)
-{
-    return !all_eq((char*)pa + osize, isize - osize, 0); // TODO: assumes little-endian
-}
-cvt_iintrinsic_checked(LLVMTrunc, check_trunc_uint, checked_trunc_uint)
-
-JL_DLLEXPORT jl_value_t *jl_check_top_bit(jl_value_t *a)
-{
-    jl_value_t *ty = jl_typeof(a);
-    if (!jl_is_primitivetype(ty))
-        jl_error("check_top_bit: value is not a primitive type");
-    if (signbitbyte(jl_data_ptr(a), jl_datatype_size(ty)))
-        jl_throw(jl_inexact_exception);
-    return a;
-}
+            jl_error("fptrunc: runtime floating point intrinsics are not implemented for bit sizes other than 32 and 64");
+#define fpext(pr, a) \
+        if (!(osize > 8 * sizeof(a))) \
+            jl_error("fpext: output bitsize must be > input bitsize"); \
+        if (osize == 32) \
+            *(float*)pr = a; \
+        else if (osize == 64) \
+            *(double*)pr = a; \
+        else \
+            jl_error("fpext: runtime floating point intrinsics are not implemented for bit sizes other than 32 and 64");
+un_fintrinsic_withtype(fptrunc,fptrunc)
+un_fintrinsic_withtype(fpext,fpext)
 
 // checked arithmetic
+/**
+ * s_typemin = - s_typemax - 1
+ * s_typemax = ((typeof a)1 << (runtime_nbits - 1)) - 1
+ * u_typemin = 0
+ * u_typemax = ((typeof a)1 << runtime_nbits) - 1
+ * where (a - a) == (typeof(a)0
+ **/
+#define sTYPEMIN(a) -sTYPEMAX(a) - 1
+#define sTYPEMAX(a)                                                \
+    (8 * sizeof(a) == runtime_nbits                                \
+         ? (((((a - a + 1) << (8 * sizeof(a) - 2)) - 1) << 1) + 1) \
+         : (  ((a - a + 1) << (runtime_nbits - 1)) - 1))
+
+#define uTYPEMIN(a) (0)
+#define uTYPEMAX(a) \
+    (8 * sizeof(~a) == runtime_nbits \
+     ? (~(a - a)) \
+     : (~((~(a - a)) << runtime_nbits)))
 #define check_sadd_int(a,b) \
-        /* this test is a reduction of (b > 0) ? (a + b > typemax(a)) : (a + b < typemin(a)) ==> overflow \
-         * where (a - a) == (typeof(a))0 */ \
-        (b > 0) ? (a > ~((a - a + 1) << (8 * sizeof(a) - 1)) - b) : (a < ((a - a + 1) << (8 * sizeof(a) - 1)) - b)
+        /* this test checks for (b >= 0) ? (a + b > typemax) : (a + b < typemin) ==> overflow */ \
+        (b >= 0) ? (a > sTYPEMAX(a) - b) : (a < sTYPEMIN(a) - b)
 checked_iintrinsic_fast(LLVMAdd_sov, check_sadd_int, add, checked_sadd_int,  )
 #define check_uadd_int(a,b) \
         /* this test checks for (a + b) > typemax(a) ==> overflow */ \
-        a >= -b
+        a > uTYPEMAX(a) - b
 checked_iintrinsic_fast(LLVMAdd_uov, check_uadd_int, add, checked_uadd_int, u)
-#define check_ssub_int(a,b) check_sadd_int(a,-b)
+#define check_ssub_int(a,b) \
+        /* this test checks for (b >= 0) ? (a - b < typemin) : (a - b > typemax) ==> overflow */ \
+        (b >= 0) ? (a < sTYPEMIN(a) + b) : (a > sTYPEMAX(a) + b)
 checked_iintrinsic_fast(LLVMSub_sov, check_ssub_int, sub, checked_ssub_int,  )
 #define check_usub_int(a,b) \
-        /* this test checks for (a - b) < 0 ==> overflow */ \
-        a < b
+        /* this test checks for (a - b) < typemin ==> overflow */ \
+        a < uTYPEMIN(a) + b
 checked_iintrinsic_fast(LLVMSub_uov, check_usub_int, sub, checked_usub_int, u)
 checked_iintrinsic_slow(LLVMMul_sov, checked_smul_int,  )
 checked_iintrinsic_slow(LLVMMul_uov, checked_umul_int, u)
@@ -905,17 +899,13 @@ checked_iintrinsic_div(LLVMRem_uov, checked_urem_int, u)
 #define flipsign(a, b) \
         (b >= 0) ? a : -a
 bi_iintrinsic_fast(jl_LLVMFlipSign, flipsign, flipsign_int,  )
-#define abs_float(pr, a) *pr = fp_select(a, fabs)
-#define ceil_float(pr, a) *pr = fp_select(a, ceil)
-#define floor_float(pr, a) *pr = fp_select(a, floor)
-#define trunc_float(pr, a) *pr = fp_select(a, trunc)
-#define rint_float(pr, a) *pr = fp_select(a, rint)
-#define sqrt_float(pr, a) \
-        if (a < 0) \
-            jl_throw(jl_domain_exception); \
-        *pr = fp_select(a, sqrt)
-#define copysign_float(a, b) \
-        fp_select2(a, b, copysign)
+#define abs_float(pr, a)      *pr = fp_select(a, fabs)
+#define ceil_float(pr, a)     *pr = fp_select(a, ceil)
+#define floor_float(pr, a)    *pr = fp_select(a, floor)
+#define trunc_float(pr, a)    *pr = fp_select(a, trunc)
+#define rint_float(pr, a)     *pr = fp_select(a, rint)
+#define sqrt_float(pr, a)     *pr = fp_select(a, sqrt)
+#define copysign_float(a, b)  fp_select2(a, b, copysign)
 
 un_fintrinsic(abs_float,abs_float)
 bi_fintrinsic(copysign_float,copysign_float)
@@ -925,12 +915,6 @@ un_fintrinsic(trunc_float,trunc_llvm)
 un_fintrinsic(rint_float,rint_llvm)
 un_fintrinsic(sqrt_float,sqrt_llvm)
 
-JL_DLLEXPORT jl_value_t *jl_select_value(jl_value_t *isfalse, jl_value_t *a, jl_value_t *b)
-{
-    JL_TYPECHK(isfalse, bool, isfalse);
-    return (isfalse == jl_false ? b : a);
-}
-
 JL_DLLEXPORT jl_value_t *jl_arraylen(jl_value_t *a)
 {
     JL_TYPECHK(arraylen, array, a);
