@@ -59,19 +59,25 @@ volatile int jl_in_stackwalk = 0;
 
 #define ROOT_TASK_STACK_ADJUSTMENT 3000000
 
+#ifdef JL_HAVE_ASYNCIFY
+// Switching logic is implemented in JavaScript
+#define STATIC_OR_JS JL_DLLEXPORT
+#else
+#define STATIC_OR_JS static
+#endif
+
 jl_sym_t *done_sym;
 jl_sym_t *failed_sym;
 jl_sym_t *runnable_sym;
 
 extern size_t jl_page_size;
-jl_datatype_t *jl_task_type;
-static char *jl_alloc_fiber(jl_ucontext_t *t, size_t *ssize, jl_task_t *owner);
-static void jl_set_fiber(jl_ucontext_t *t);
-static void jl_start_fiber(jl_ucontext_t *lastt, jl_ucontext_t *t);
-static void jl_swap_fiber(jl_ucontext_t *lastt, jl_ucontext_t *t);
+static char *jl_alloc_fiber(jl_ucontext_t *t, size_t *ssize, jl_task_t *owner) JL_NOTSAFEPOINT;
+STATIC_OR_JS void jl_set_fiber(jl_ucontext_t *t);
+STATIC_OR_JS void jl_start_fiber(jl_ucontext_t *lastt, jl_ucontext_t *t);
+STATIC_OR_JS void jl_swap_fiber(jl_ucontext_t *lastt, jl_ucontext_t *t);
 
 #ifdef JL_HAVE_UNW_CONTEXT
-static JL_THREAD unw_cursor_t jl_basecursor;
+static JL_THREAD_LOCAL unw_cursor_t jl_basecursor;
 #endif
 
 #ifdef ALWAYS_COPY_STACKS
@@ -118,9 +124,8 @@ static void NOINLINE save_stack(jl_ptls_t ptls, jl_task_t *lastt, jl_task_t **pt
     jl_gc_wb_back(lastt);
 }
 
-static void NOINLINE JL_NORETURN restore_stack(jl_ptls_t ptls, char *p)
+static void NOINLINE JL_NORETURN restore_stack(jl_task_t *t, jl_ptls_t ptls, char *p)
 {
-    jl_task_t *t = ptls->current_task;
     size_t nb = t->copy_stack;
     char *_x = (char*)ptls->stackbase - nb;
     if (!p) {
@@ -129,16 +134,15 @@ static void NOINLINE JL_NORETURN restore_stack(jl_ptls_t ptls, char *p)
         if ((char*)&_x > _x) {
             p = (char*)alloca((char*)&_x - _x);
         }
-        restore_stack(ptls, p); // pass p to ensure the compiler can't tailcall this or avoid the alloca
+        restore_stack(t, ptls, p); // pass p to ensure the compiler can't tailcall this or avoid the alloca
     }
     assert(t->stkbuf != NULL);
     memcpy_a16((uint64_t*)_x, (uint64_t*)t->stkbuf, nb); // destroys all but the current stackframe
     jl_set_fiber(&t->ctx);
     abort(); // unreachable
 }
-static void restore_stack2(jl_ptls_t ptls, jl_task_t *lastt)
+static void restore_stack2(jl_task_t *t, jl_ptls_t ptls, jl_task_t *lastt)
 {
-    jl_task_t *t = ptls->current_task;
     size_t nb = t->copy_stack;
     char *_x = (char*)ptls->stackbase - nb;
     assert(t->stkbuf != NULL);
@@ -147,20 +151,21 @@ static void restore_stack2(jl_ptls_t ptls, jl_task_t *lastt)
 }
 #endif
 
-static jl_function_t *task_done_hook_func = NULL;
+/* Rooted by the base module */
+static jl_function_t *task_done_hook_func JL_GLOBALLY_ROOTED = NULL;
 
 void JL_NORETURN jl_finish_task(jl_task_t *t, jl_value_t *resultval JL_MAYBE_UNROOTED)
 {
     jl_ptls_t ptls = jl_get_ptls_states();
     JL_SIGATOMIC_BEGIN();
-    if (t->exception != jl_nothing)
-        t->state = failed_sym;
-    else
-        t->state = done_sym;
-    if (t->copy_stack) // early free of stkbuf
-        t->stkbuf = NULL;
     t->result = resultval;
     jl_gc_wb(t, t->result);
+    if (t->exception != jl_nothing)
+        jl_atomic_store_release(&t->state, failed_sym);
+    else
+        jl_atomic_store_release(&t->state, done_sym);
+    if (t->copy_stack) // early free of stkbuf
+        t->stkbuf = NULL;
     // ensure that state is cleared
     ptls->in_finalizer = 0;
     ptls->in_pure_callback = 0;
@@ -231,12 +236,6 @@ static void ctx_switch(jl_ptls_t ptls, jl_task_t **pt)
     jl_task_t *t = *pt;
     assert(t != ptls->current_task);
     jl_task_t *lastt = ptls->current_task;
-#ifdef ENABLE_TIMINGS
-    jl_timing_block_t *blk = lastt->timing_stack;
-    if (blk)
-        jl_timing_block_stop(blk);
-#endif
-#ifdef JULIA_ENABLE_THREADING
     // If the current task is not holding any locks, free the locks list
     // so that it can be GC'd without leaking memory
     arraylist_t *locks = &lastt->locks;
@@ -244,7 +243,6 @@ static void ctx_switch(jl_ptls_t ptls, jl_task_t **pt)
         arraylist_free(locks);
         arraylist_new(locks, 0);
     }
-#endif
 
     int started = t->started;
     int killed = (lastt->state == done_sym || lastt->state == failed_sym);
@@ -280,11 +278,6 @@ static void ctx_switch(jl_ptls_t ptls, jl_task_t **pt)
             save_stack(ptls, lastt, pt); // allocates (gc-safepoint, and can also fail)
             if (jl_setjmp(lastt->ctx.uc_mcontext, 0)) {
                 // TODO: mutex unlock the thread we just switched from
-#ifdef ENABLE_TIMINGS
-                assert(blk == ptls->current_task->timing_stack);
-                if (blk)
-                    jl_timing_block_start(blk);
-#endif
                 return;
             }
         }
@@ -299,11 +292,10 @@ static void ctx_switch(jl_ptls_t ptls, jl_task_t **pt)
     ptls->pgcstack = t->gcstack;
     ptls->world_age = t->world_age;
     t->gcstack = NULL;
+#ifdef MIGRATE_TASKS
+    ptls->previous_task = lastt;
+#endif
     ptls->current_task = t;
-    if (!lastt->sticky)
-        // release lastt to run on any tid
-        lastt->tid = -1;
-    t->tid = ptls->tid;
 
     jl_ucontext_t *lastt_ctx = (killed ? NULL : &lastt->ctx);
 #ifdef COPY_STACKS
@@ -318,11 +310,11 @@ static void ctx_switch(jl_ptls_t ptls, jl_task_t **pt)
 #ifdef COPY_STACKS
         if (t->copy_stack) {
             if (lastt_ctx)
-                restore_stack2(ptls, lastt);
+                restore_stack2(t, ptls, lastt);
             else if (lastt->copy_stack)
-                restore_stack(ptls, NULL);     // (doesn't return)
+                restore_stack(t, ptls, NULL);     // (doesn't return)
             else
-                restore_stack(ptls, (char*)1); // (doesn't return)
+                restore_stack(t, ptls, (char*)1); // (doesn't return)
         }
         else
 #endif
@@ -332,39 +324,79 @@ static void ctx_switch(jl_ptls_t ptls, jl_task_t **pt)
             jl_swap_fiber(lastt_ctx, &t->ctx);
     }
     else {
+#ifdef COPY_STACKS
         if (always_copy_stacks)
             jl_longjmp(ptls->base_ctx.uc_mcontext, 1);
         else
+#endif
             jl_start_fiber(lastt_ctx, &t->ctx);
     }
-    // TODO: mutex unlock the thread we just switched from
-#ifdef ENABLE_TIMINGS
-    assert(blk == ptls->current_task->timing_stack);
-    if (blk)
-        jl_timing_block_start(blk);
-#endif
+}
+
+static jl_ptls_t NOINLINE refetch_ptls(void)
+{
+    return jl_get_ptls_states();
 }
 
 JL_DLLEXPORT void jl_switchto(jl_task_t **pt)
 {
     jl_ptls_t ptls = jl_get_ptls_states();
     jl_task_t *t = *pt;
-    if (t == ptls->current_task) {
+    jl_task_t *ct = ptls->current_task;
+    if (t == ct) {
         return;
     }
     if (t->state == done_sym || t->state == failed_sym ||
             (t->started && t->stkbuf == NULL)) {
-        ptls->current_task->exception = t->exception;
-        ptls->current_task->result = t->result;
+        ct->exception = t->exception;
+        ct->result = t->result;
         return;
     }
     if (ptls->in_finalizer)
         jl_error("task switch not allowed from inside gc finalizer");
     if (ptls->in_pure_callback)
         jl_error("task switch not allowed from inside staged nor pure functions");
+    if (t->sticky && jl_atomic_load_acquire(&t->tid) == -1) {
+        // manually yielding to a task
+        if (jl_atomic_compare_exchange(&t->tid, -1, ptls->tid) != -1)
+            jl_error("cannot switch to task running on another thread");
+    }
+    else if (t->tid != ptls->tid) {
+        jl_error("cannot switch to task running on another thread");
+    }
     sig_atomic_t defer_signal = ptls->defer_signal;
     int8_t gc_state = jl_gc_unsafe_enter(ptls);
+
+#ifdef ENABLE_TIMINGS
+    jl_timing_block_t *blk = ct->timing_stack;
+    if (blk)
+        jl_timing_block_stop(blk);
+#endif
+
     ctx_switch(ptls, pt);
+
+#ifdef MIGRATE_TASKS
+    ptls = refetch_ptls();
+    t = ptls->previous_task;
+    assert(t->tid == ptls->tid);
+    if (!t->sticky && !t->copy_stack)
+        t->tid = -1;
+#elif defined(NDEBUG)
+    (void)refetch_ptls();
+#else
+    assert(ptls == refetch_ptls());
+#endif
+
+    ct = ptls->current_task;
+
+#ifdef ENABLE_TIMINGS
+    assert(blk == ct->timing_stack);
+    if (blk)
+        jl_timing_block_start(blk);
+#else
+    (void)ct;
+#endif
+
     jl_gc_unsafe_leave(ptls, gc_state);
     sig_atomic_t other_defer_signal = ptls->defer_signal;
     ptls->defer_signal = defer_signal;
@@ -372,7 +404,7 @@ JL_DLLEXPORT void jl_switchto(jl_task_t **pt)
         jl_sigint_safepoint(ptls);
 }
 
-JL_DLLEXPORT JL_NORETURN void jl_no_exc_handler(jl_value_t *e) JL_NOTSAFEPOINT
+JL_DLLEXPORT JL_NORETURN void jl_no_exc_handler(jl_value_t *e)
 {
     jl_printf(JL_STDERR, "fatal: error thrown and no exception handler available.\n");
     jl_static_show(JL_STDERR, e);
@@ -507,13 +539,11 @@ JL_DLLEXPORT jl_task_t *jl_new_task(jl_function_t *start, jl_value_t *completion
     t->stkbuf = NULL;
     t->started = 0;
     t->prio = -1;
-    t->tid = 0;
+    t->tid = -1;
 #ifdef ENABLE_TIMINGS
     t->timing_stack = NULL;
 #endif
-#ifdef JULIA_ENABLE_THREADING
     arraylist_new(&t->locks, 0);
-#endif
 
 #if defined(JL_DEBUG_BUILD)
     if (!t->copy_stack)
@@ -532,52 +562,85 @@ JL_DLLEXPORT jl_value_t *jl_get_current_task(void)
     return (jl_value_t*)ptls->current_task;
 }
 
+#ifdef JL_HAVE_ASYNCIFY
+JL_DLLEXPORT jl_ucontext_t *task_ctx_ptr(jl_task_t *t)
+{
+    return &t->ctx;
+}
+
+JL_DLLEXPORT jl_value_t *jl_get_root_task(void)
+{
+    jl_ptls_t ptls = jl_get_ptls_states();
+    return (jl_value_t*)ptls->root_task;
+}
+
+void JL_DLLEXPORT jl_task_wait()
+{
+    static jl_function_t *wait_func = NULL;
+    if (!wait_func) {
+        wait_func = (jl_function_t*)jl_get_global(jl_base_module, jl_symbol("wait"));
+    }
+    size_t last_age = jl_get_ptls_states()->world_age;
+    jl_get_ptls_states()->world_age = jl_get_world_counter();
+    jl_apply(&wait_func, 1);
+    jl_get_ptls_states()->world_age = last_age;
+}
+
+void JL_DLLEXPORT jl_schedule_task(jl_task_t *task)
+{
+    static jl_function_t *sched_func = NULL;
+    if (!sched_func) {
+        sched_func = (jl_function_t*)jl_get_global(jl_base_module, jl_symbol("schedule"));
+    }
+    size_t last_age = jl_get_ptls_states()->world_age;
+    jl_get_ptls_states()->world_age = jl_get_world_counter();
+    jl_value_t *args[] = {(jl_value_t*)sched_func, (jl_value_t*)task};
+    jl_apply(args, 2);
+    jl_get_ptls_states()->world_age = last_age;
+}
+#endif
+
 // Do one-time initializations for task system
 void jl_init_tasks(void) JL_GC_DISABLED
 {
-    jl_task_type = (jl_datatype_t*)
-        jl_new_datatype(jl_symbol("Task"),
-                        NULL,
-                        jl_any_type,
-                        jl_emptysvec,
-                        jl_perm_symsvec(11,
-                                        "next",
-                                        "queue",
-                                        "storage",
-                                        "state",
-                                        "donenotify",
-                                        "result",
-                                        "exception",
-                                        "backtrace",
-                                        "logstate",
-                                        "code",
-                                        "sticky"),
-                        jl_svec(11,
-                                jl_any_type,
-                                jl_any_type,
-                                jl_any_type,
-                                jl_sym_type,
-                                jl_any_type,
-                                jl_any_type,
-                                jl_any_type,
-                                jl_any_type,
-                                jl_any_type,
-                                jl_any_type,
-                                jl_bool_type),
-                        0, 1, 9);
-    jl_value_t *listt = jl_new_struct(jl_uniontype_type, jl_task_type, jl_void_type);
-    jl_svecset(jl_task_type->types, 0, listt);
     done_sym = jl_symbol("done");
     failed_sym = jl_symbol("failed");
     runnable_sym = jl_symbol("runnable");
+
+    char *acs = getenv("JULIA_COPY_STACKS");
+    if (acs) {
+        if (!strcmp(acs, "1") || !strcmp(acs, "yes"))
+            always_copy_stacks = 1;
+        else if (!strcmp(acs, "0") || !strcmp(acs, "no"))
+            always_copy_stacks = 0;
+        else {
+            jl_printf(JL_STDERR, "invalid JULIA_COPY_STACKS value: %s\n", acs);
+            exit(1);
+        }
+    }
 }
 
-static void NOINLINE JL_NORETURN start_task(void)
+STATIC_OR_JS void NOINLINE JL_NORETURN start_task(void)
 {
+#ifdef _OS_WINDOWS_
+#if defined(_CPU_X86_64_)
+    // install the unhandled exception hanlder at the top of our stack
+    // to call directly into our personality handler
+    asm volatile ("\t.seh_handler __julia_personality, @except\n\t.text");
+#endif
+#endif
+
     // this runs the first time we switch to a task
     jl_ptls_t ptls = jl_get_ptls_states();
     jl_task_t *t = ptls->current_task;
     jl_value_t *res;
+
+#ifdef MIGRATE_TASKS
+    jl_task_t *pt = ptls->previous_task;
+    if (!pt->sticky && !pt->copy_stack)
+        pt->tid = -1;
+#endif
+
     t->started = 1;
     if (t->exception != jl_nothing) {
         record_backtrace(ptls);
@@ -616,7 +679,7 @@ skip_pop_exception:;
 #define swapcontext jl_swapcontext
 #define makecontext jl_makecontext
 #endif
-static char *jl_alloc_fiber(jl_ucontext_t *t, size_t *ssize, jl_task_t *owner)
+static char *jl_alloc_fiber(jl_ucontext_t *t, size_t *ssize, jl_task_t *owner) JL_NOTSAFEPOINT
 {
 #ifndef _OS_WINDOWS_
     int r = getcontext(t);
@@ -743,8 +806,10 @@ static char *jl_alloc_fiber(jl_ucontext_t *t, size_t *ssize, jl_task_t *owner)
     char *stkbuf = (char*)jl_malloc_stack(ssize, owner);
     if (stkbuf == NULL)
         return NULL;
+#ifndef __clang_analyzer__
     ((char**)t)[0] = stkbuf; // stash the stack pointer somewhere for start_fiber
     ((size_t*)t)[1] = *ssize; // stash the stack size somewhere for start_fiber
+#endif
     return stkbuf;
 }
 static void jl_start_fiber(jl_ucontext_t *lastt, jl_ucontext_t *t)
@@ -929,12 +994,32 @@ static void jl_init_basefiber(size_t ssize)
 }
 #endif
 
+#if defined(JL_HAVE_ASYNCIFY)
+static void jl_init_basefiber(size_t ssize)
+{
+}
+static char *jl_alloc_fiber(jl_ucontext_t *t, size_t *ssize, jl_task_t *owner) JL_NOTSAFEPOINT
+{
+    void *stk = jl_malloc_stack(ssize, owner);
+    if (stk == NULL)
+        return NULL;
+    t->stackbottom = stk;
+    t->stacktop = ((char*)stk) + *ssize;
+    return (char*)stk;
+}
+// jl_*_fiber implemented in js
+#endif
+
 // Initialize a root task using the given stack.
 void jl_init_root_task(void *stack_lo, void *stack_hi)
 {
     jl_ptls_t ptls = jl_get_ptls_states();
-    ptls->current_task = (jl_task_t*)jl_gc_alloc(ptls, sizeof(jl_task_t),
-                                                 jl_task_type);
+    if (ptls->root_task == NULL) {
+        ptls->root_task = (jl_task_t*)jl_gc_alloc(ptls, sizeof(jl_task_t), jl_task_type);
+        memset(ptls->root_task, 0, sizeof(jl_task_t));
+        ptls->root_task->tls = jl_nothing;
+    }
+    ptls->current_task = ptls->root_task;
     void *stack = stack_lo;
     size_t ssize = (char*)stack_hi - (char*)stack_lo;
 #ifndef _OS_WINDOWS_
@@ -956,7 +1041,6 @@ void jl_init_root_task(void *stack_lo, void *stack_hi)
     ptls->current_task->started = 1;
     ptls->current_task->next = jl_nothing;
     ptls->current_task->queue = jl_nothing;
-    ptls->current_task->tls = jl_nothing;
     ptls->current_task->state = runnable_sym;
     ptls->current_task->start = NULL;
     ptls->current_task->result = jl_nothing;
@@ -969,21 +1053,19 @@ void jl_init_root_task(void *stack_lo, void *stack_hi)
     ptls->current_task->excstack = NULL;
     ptls->current_task->tid = ptls->tid;
     ptls->current_task->sticky = 1;
-#ifdef JULIA_ENABLE_THREADING
     arraylist_new(&ptls->current_task->locks, 0);
-#endif
-
-    ptls->root_task = ptls->current_task;
 
+#ifdef COPY_STACKS
     if (always_copy_stacks) {
         ptls->stackbase = stack_hi;
         ptls->stacksize = ssize;
         if (jl_setjmp(ptls->base_ctx.uc_mcontext, 0))
             start_task();
+        return;
     }
-    else {
-        jl_init_basefiber(JL_STACK_SIZE);
-    }
+#endif
+
+    jl_init_basefiber(JL_STACK_SIZE);
 }
 
 JL_DLLEXPORT int jl_is_task_started(jl_task_t *t)
