@@ -49,8 +49,7 @@ extern "C" {
 // as the fallback in the shared object. For better efficiency, we also
 // create a `__thread` variable in the main executable using a static TLS
 // model.
-#ifdef JULIA_ENABLE_THREADING
-#  if defined(_OS_DARWIN_)
+#if defined(_OS_DARWIN_)
 // Mac doesn't seem to have static TLS model so the runtime TLS getter
 // registration will only add overhead to TLS access. The `__thread` variables
 // are emulated with `pthread_key_t` so it is actually faster to use it directly.
@@ -61,7 +60,7 @@ __attribute__((constructor)) void jl_mac_init_tls(void)
     pthread_key_create(&jl_tls_key, NULL);
 }
 
-JL_DLLEXPORT JL_CONST_FUNC jl_ptls_t (jl_get_ptls_states)(void)
+JL_DLLEXPORT JL_CONST_FUNC jl_ptls_t (jl_get_ptls_states)(void) JL_GLOBALLY_ROOTED
 {
     void *ptls = pthread_getspecific(jl_tls_key);
     if (__unlikely(!ptls)) {
@@ -82,7 +81,7 @@ jl_get_ptls_states_func jl_get_ptls_states_getter(void)
     // for codegen
     return &jl_get_ptls_states_fast;
 }
-#  elif defined(_OS_WINDOWS_)
+#elif defined(_OS_WINDOWS_)
 // Apparently windows doesn't have a static TLS model (or one that can be
 // reliably used from a shared library) either..... Use `TLSAlloc` instead.
 
@@ -112,7 +111,7 @@ BOOLEAN WINAPI DllMain(IN HINSTANCE hDllHandle, IN DWORD nReason,
     return 1; // success
 }
 
-JL_DLLEXPORT JL_CONST_FUNC jl_ptls_t (jl_get_ptls_states)(void)
+JL_DLLEXPORT JL_CONST_FUNC jl_ptls_t (jl_get_ptls_states)(void) JL_GLOBALLY_ROOTED
 {
     return (jl_ptls_t)TlsGetValue(jl_tls_key);
 }
@@ -122,7 +121,7 @@ jl_get_ptls_states_func jl_get_ptls_states_getter(void)
     // for codegen
     return &jl_get_ptls_states;
 }
-#  else
+#else
 // We use the faster static version in the main executable to replace
 // the slower version in the shared object. The code in different libraries
 // or executables, however, have to agree on which version to use.
@@ -143,10 +142,10 @@ static JL_CONST_FUNC jl_ptls_t jl_get_ptls_states_fallback(void)
     static __thread jl_tls_states_t tls_states;
     return &tls_states;
 }
-#if JL_USE_IFUNC
+#  if JL_USE_IFUNC
 JL_DLLEXPORT JL_CONST_FUNC __attribute__((weak))
 jl_ptls_t jl_get_ptls_states_static(void);
-#endif
+#  endif
 static jl_ptls_t jl_get_ptls_states_init(void);
 static jl_get_ptls_states_func jl_tls_states_cb = jl_get_ptls_states_init;
 static jl_ptls_t jl_get_ptls_states_init(void)
@@ -160,17 +159,19 @@ static jl_ptls_t jl_get_ptls_states_init(void)
     // make sure the tls states callback is finalized before adding
     // multiple threads
     jl_get_ptls_states_func cb = jl_get_ptls_states_fallback;
-#if JL_USE_IFUNC
+#  if JL_USE_IFUNC
     if (jl_get_ptls_states_static)
         cb = jl_get_ptls_states_static;
-#endif
+#  endif
     jl_tls_states_cb = cb;
     return cb();
 }
 
-static JL_CONST_FUNC jl_ptls_t jl_get_ptls_states_wrapper(void)
+static JL_CONST_FUNC jl_ptls_t jl_get_ptls_states_wrapper(void) JL_GLOBALLY_ROOTED JL_NOTSAFEPOINT
 {
+#ifndef __clang_analyzer__
     return (*jl_tls_states_cb)();
+#endif
 }
 
 JL_DLLEXPORT void jl_set_ptls_states_getter(jl_get_ptls_states_func f)
@@ -187,7 +188,7 @@ JL_DLLEXPORT void jl_set_ptls_states_getter(jl_get_ptls_states_func f)
     }
 }
 
-#if JL_USE_IFUNC
+#  if JL_USE_IFUNC
 static jl_get_ptls_states_func jl_get_ptls_states_resolve(void)
 {
     if (jl_tls_states_cb != jl_get_ptls_states_init)
@@ -202,14 +203,14 @@ static jl_get_ptls_states_func jl_get_ptls_states_resolve(void)
     return jl_tls_states_cb;
 }
 
-JL_DLLEXPORT JL_CONST_FUNC jl_ptls_t (jl_get_ptls_states)(void)
+JL_DLLEXPORT JL_CONST_FUNC jl_ptls_t (jl_get_ptls_states)(void) JL_GLOBALLY_ROOTED
     __attribute__((ifunc ("jl_get_ptls_states_resolve")));
-#else // JL_TLS_USE_IFUNC
-JL_DLLEXPORT JL_CONST_FUNC jl_ptls_t (jl_get_ptls_states)(void)
+#  else // JL_TLS_USE_IFUNC
+JL_DLLEXPORT JL_CONST_FUNC jl_ptls_t (jl_get_ptls_states)(void) JL_GLOBALLY_ROOTED
 {
     return jl_get_ptls_states_wrapper();
 }
-#endif // JL_TLS_USE_IFUNC
+#  endif // JL_TLS_USE_IFUNC
 jl_get_ptls_states_func jl_get_ptls_states_getter(void)
 {
     if (jl_tls_states_cb == jl_get_ptls_states_init)
@@ -217,17 +218,10 @@ jl_get_ptls_states_func jl_get_ptls_states_getter(void)
     // for codegen
     return jl_tls_states_cb;
 }
-#  endif
-#else
-JL_DLLEXPORT jl_tls_states_t jl_tls_states;
-JL_DLLEXPORT JL_CONST_FUNC jl_ptls_t (jl_get_ptls_states)(void)
-{
-    return &jl_tls_states;
-}
 #endif
 
 JL_DLLEXPORT int jl_n_threads;
-jl_ptls_t *jl_all_tls_states;
+jl_ptls_t *jl_all_tls_states JL_GLOBALLY_ROOTED;
 
 // return calling thread's ID
 // Also update the suspended_threads list in signals-mach when changing the
@@ -241,6 +235,7 @@ JL_DLLEXPORT int16_t jl_threadid(void)
 void jl_init_threadtls(int16_t tid)
 {
     jl_ptls_t ptls = jl_get_ptls_states();
+    ptls->system_id = jl_thread_self();
     seed_cong(&ptls->rngseed);
 #ifdef _OS_WINDOWS_
     if (tid == 0) {
@@ -251,8 +246,6 @@ void jl_init_threadtls(int16_t tid)
             hMainThread = INVALID_HANDLE_VALUE;
         }
     }
-#else
-    ptls->system_id = pthread_self();
 #endif
     assert(ptls->world_age == 0);
     ptls->world_age = 1; // OK to run Julia code on this thread
@@ -292,8 +285,6 @@ void jl_init_threadtls(int16_t tid)
 jl_mutex_t codegen_lock;
 jl_mutex_t typecache_lock;
 
-#ifdef JULIA_ENABLE_THREADING
-
 ssize_t jl_tls_offset = -1;
 
 #ifdef JL_ELF_TLS_VARIANT
@@ -410,9 +401,9 @@ void jl_init_threading(void)
         jl_n_threads = max_threads;
     if (jl_n_threads <= 0)
         jl_n_threads = 1;
-
+#ifndef __clang_analyzer__
     jl_all_tls_states = (jl_ptls_t*)calloc(jl_n_threads, sizeof(void*));
-
+#endif
     // initialize this thread (set tid, create heap, etc.)
     jl_init_threadtls(0);
 
@@ -471,9 +462,7 @@ void jl_start_threads(void)
     uv_barrier_wait(&thread_init_done);
 }
 
-#endif
-
-unsigned volatile _threadedregion; // HACK: prevent the root task from sleeping
+unsigned volatile _threadedregion; // HACK: keep track of whether it is safe to do IO
 
 // simple fork/join mode code
 JL_DLLEXPORT void jl_threading_run(jl_value_t *func)
@@ -505,23 +494,15 @@ JL_DLLEXPORT void jl_threading_run(jl_value_t *func)
         args2[0] = schd_func;
         args2[1] = (jl_value_t*)t;
         jl_apply(args2, 2);
-#ifdef JULIA_ENABLE_THREADING
         if (i == 1) {
             // let threads know work is coming (optimistic)
-            uv_mutex_lock(&sleep_lock);
-            uv_cond_broadcast(&sleep_alarm);
-            uv_mutex_unlock(&sleep_lock);
+            jl_wakeup_thread(-1);
         }
-#endif
     }
-#ifdef JULIA_ENABLE_THREADING
     if (nthreads > 2) {
         // let threads know work is ready (guaranteed)
-        uv_mutex_lock(&sleep_lock);
-        uv_cond_broadcast(&sleep_alarm);
-        uv_mutex_unlock(&sleep_lock);
+        jl_wakeup_thread(-1);
     }
-#endif
     // join with all tasks
     JL_TRY {
         for (int i = 0; i < nthreads; i++) {
@@ -537,8 +518,8 @@ JL_DLLEXPORT void jl_threading_run(jl_value_t *func)
         JL_UV_UNLOCK();
         jl_rethrow();
     }
-    _threadedregion -= 1;
     // make sure no threads are sitting in the event loop
+    _threadedregion -= 1;
     jl_wake_libuv();
     // make sure no more callbacks will run while user code continues
     // outside thread region and might touch an I/O object.
@@ -549,20 +530,6 @@ JL_DLLEXPORT void jl_threading_run(jl_value_t *func)
 }
 
 
-#ifndef JULIA_ENABLE_THREADING
-
-void jl_init_threading(void)
-{
-    static jl_ptls_t _jl_all_tls_states;
-    jl_all_tls_states = &_jl_all_tls_states;
-    jl_n_threads = 1;
-    jl_init_threadtls(0);
-}
-
-void jl_start_threads(void) { }
-
-#endif // !JULIA_ENABLE_THREADING
-
 // Make gc alignment available for threading
 // see threads.jl alignment
 JL_DLLEXPORT int jl_alignment(size_t sz)
