@@ -425,6 +425,45 @@ end
     end
 end
 
+@testset "larfg & larf" begin
+    @testset for elty in (Float32, Float64, ComplexF32, ComplexF64)
+        ## larfg
+        Random.seed!(0)
+        x  = rand(elty, 5)
+        v  = copy(x)
+        τ = LinearAlgebra.LAPACK.larfg!(v)
+        H = (I - τ*v*v')
+        # for complex input, LAPACK wants a conjugate transpose of H (check clarfg docs)
+        y = elty <: Complex ? H'*x : H*x
+        # we have rotated a vector
+        @test norm(y) ≈ norm(x)
+        # an annihilated almost all the first column
+        @test norm(y[2:end], Inf) < 10*eps(real(one(elty)))
+
+        ## larf
+        C = rand(elty, 5, 5)
+        C_norm = norm(C, 2)
+        v = C[1:end, 1]
+        τ = LinearAlgebra.LAPACK.larfg!(v)
+        LinearAlgebra.LAPACK.larf!('L', v, conj(τ), C)
+        # we have applied a unitary transformation
+        @test norm(C, 2) ≈ C_norm
+        # an annihilated almost all the first column
+        @test norm(C[2:end, 1], Inf) < 10*eps(real(one(elty)))
+
+        # apply left and right
+        C1 = rand(elty, 5, 5)
+        C2 = rand(elty, 5, 5)
+        C = C2*C1
+
+        v = C1[1:end, 1]
+        τ = LinearAlgebra.LAPACK.larfg!(v)
+        LinearAlgebra.LAPACK.larf!('L', v,      τ, C1)
+        LinearAlgebra.LAPACK.larf!('R', v, conj(τ), C2)
+        @test C ≈ C2*C1
+    end
+end
+
 @testset "tgsen, tzrzf, & trsyl" begin
     @testset for elty in (Float32, Float64, ComplexF32, ComplexF64)
         Z = zeros(elty,10,10)
