@@ -1,11 +1,45 @@
 # This file is a part of Julia. License is MIT: https://julialang.org/license
 
 import Base.copy, Base.==
+using Random
+using InteractiveUtils: code_llvm
+
+import Libdl
 
 const libccalltest = "libccalltest"
 
 const verbose = false
-ccall((:set_verbose, libccalltest), Void, (Int32,), verbose)
+ccall((:set_verbose, libccalltest), Cvoid, (Int32,), verbose)
+
+@eval function cvarargs()
+    strp = Ref{Ptr{Cchar}}(0)
+    fmt = "%3.1f"
+    len = ccall(:asprintf, Cint, (Ptr{Ptr{Cchar}}, Cstring, Cfloat...), strp, fmt, 0.1)
+    str = unsafe_string(strp[], len)
+    Libc.free(strp[])
+    return str
+end
+@test cvarargs() == "0.1"
+
+
+# test multiple-type vararg handling (there's no syntax for this currently)
+@eval function foreign_varargs()
+    strp = Ref{Ptr{Cchar}}(0)
+    fmt = "hi+%hhd-%hhd-%hhd-%hhd-%hhd-%hhd-%hhd-%hhd-%hhd-%hhd-%hhd-%hhd-%hhd-%hhd-%hhd-%.1f-%.1f-%.1f-%.1f-%.1f-%.1f-%.1f-%.1f-%.1f\n"
+    len = $(Expr(:foreigncall, :(:asprintf), Cint,
+        Core.svec(Ptr{Ptr{Cchar}}, Cstring,
+            UInt8, UInt8, UInt8, UInt8, UInt8, UInt8, UInt8, UInt8, UInt8, UInt8, UInt8, UInt8, UInt8, UInt8, UInt8,
+            Cfloat, Cfloat, Cfloat, Cfloat, Cfloat, Cfloat, Cfloat, Cfloat, Cfloat),
+            2, :(:cdecl),
+            :(Base.unsafe_convert(Ptr{Ptr{Cchar}}, strp)), :(Base.unsafe_convert(Cstring, fmt)),
+            0x1, 0x2, 0x3, 0x4, 0x5, 0x6, 0x7, 0x8, 0x9, 0xa, 0xb, 0xc, 0xd, 0xe, 0xf,
+            Cfloat(1.1), Cfloat(2.2), Cfloat(3.3), Cfloat(4.4), Cfloat(5.5), Cfloat(6.6), Cfloat(7.7), Cfloat(8.8), Cfloat(9.9),
+            :strp, :fmt))
+    str = unsafe_string(strp[], len)
+    Libc.free(strp[])
+    return str
+end
+@test foreign_varargs() == "hi+1-2-3-4-5-6-7-8-9-10-11-12-13-14-15-1.1-2.2-3.3-4.4-5.5-6.6-7.7-8.8-9.9\n"
 
 
 # Test for proper argument register truncation
@@ -82,20 +116,20 @@ let a, ci_ary, x
     @test x == a + 1 - 2im
     @test a == 20 + 51im
 
-    x = ccall((:cptest_static, libccalltest), Ptr{Complex{Int}}, (Ptr{Complex{Int}},), &a)
+    x = ccall((:cptest_static, libccalltest), Ptr{Complex{Int}}, (Ref{Complex{Int}},), a)
     @test unsafe_load(x) == a
-    Libc.free(convert(Ptr{Void}, x))
+    Libc.free(convert(Ptr{Cvoid}, x))
 end
 
 let a, b, x
     a = 2.84 + 5.2im
 
-    x = ccall((:cgtest, libccalltest), Complex128, (Complex128,), a)
+    x = ccall((:cgtest, libccalltest), ComplexF64, (ComplexF64,), a)
 
     @test x == a + 1 - 2im
 
     b = [a] # Make sure the array is alive during unsafe_load
-    x = unsafe_load(ccall((:cgptest, libccalltest), Ptr{Complex128}, (Ptr{Complex128},), b))
+    x = unsafe_load(ccall((:cgptest, libccalltest), Ptr{ComplexF64}, (Ptr{ComplexF64},), b))
 
     @test x == a + 1 - 2im
     @test a == 2.84 + 5.2im
@@ -104,12 +138,12 @@ end
 let a, b, x
     a = 3.34f0 + 53.2f0im
 
-    x = ccall((:cftest, libccalltest), Complex64, (Complex64,), a)
+    x = ccall((:cftest, libccalltest), ComplexF32, (ComplexF32,), a)
 
     @test x == a + 1 - 2im
 
     b = [a] # Make sure the array is alive during unsafe_load
-    x = unsafe_load(ccall((:cfptest, libccalltest), Ptr{Complex64}, (Ptr{Complex64},), b))
+    x = unsafe_load(ccall((:cfptest, libccalltest), Ptr{ComplexF32}, (Ptr{ComplexF32},), b))
 
     @test x == a + 1 - 2im
     @test a == 3.34f0 + 53.2f0im
@@ -122,7 +156,6 @@ let a
     a = 2.84 + 5.2im
 
     @test_throws MethodError ccall((:cptest, libccalltest), Ptr{Complex{Int}}, (Ptr{Complex{Int}},), a)
-    @test_throws MethodError ccall((:cptest, libccalltest), Ptr{Complex{Int}}, (Complex{Int},), &a)
 end
 
 
@@ -139,7 +172,7 @@ end
 copy(a::Struct1) = Struct1(a.x, a.y)
 copy(a::Struct1I) = a
 
-function test_struct1{Struct}(::Type{Struct})
+function test_struct1(::Type{Struct}) where {Struct}
     a = Struct(352.39422f23, 19.287577)
     b = Float32(123.456)
 
@@ -223,7 +256,7 @@ struct Struct4I
     z::Int32
 end
 
-function test_struct4{Struct}(::Type{Struct})
+function test_struct4(::Type{Struct}) where {Struct}
     a = Struct(-512275808,882558299,-2133022131)
     b = Int32(42)
 
@@ -253,7 +286,7 @@ struct Struct5I
     a::Int32
 end
 
-function test_struct5{Struct}(::Type{Struct})
+function test_struct5(::Type{Struct}) where {Struct}
     a = Struct(1771319039, 406394736, -1269509787, -745020976)
     b = Int32(42)
 
@@ -282,7 +315,7 @@ struct Struct6I
     z::Int64
 end
 
-function test_struct6{Struct}(::Type{Struct})
+function test_struct6(::Type{Struct}) where {Struct}
     a = Struct(-654017936452753226, -5573248801240918230, -983717165097205098)
     b = Int64(42)
 
@@ -308,7 +341,7 @@ struct Struct7I
     y::Cchar
 end
 
-function test_struct7{Struct}(::Type{Struct})
+function test_struct7(::Type{Struct}) where {Struct}
     a = Struct(-384082741977533896, 'h')
     b = Int8(42)
 
@@ -333,7 +366,7 @@ struct Struct8I
     y::Cchar
 end
 
-function test_struct8{Struct}(::Type{Struct})
+function test_struct8(::Type{Struct}) where {Struct}
     a = Struct(-384082896, 'h')
     b = Int8(42)
 
@@ -358,7 +391,7 @@ struct Struct9I
     y::Int16
 end
 
-function test_struct9{Struct}(::Type{Struct})
+function test_struct9(::Type{Struct}) where {Struct}
     a = Struct(-394092996, -3840)
     b = Int16(42)
 
@@ -387,7 +420,7 @@ struct Struct10I
     a::Cchar
 end
 
-function test_struct10{Struct}(::Type{Struct})
+function test_struct10(::Type{Struct}) where {Struct}
     a = Struct('0', '1', '2', '3')
     b = Int8(2)
 
@@ -406,13 +439,13 @@ test_struct10(Struct10)
 test_struct10(Struct10I)
 
 mutable struct Struct11
-    x::Complex64
+    x::ComplexF32
 end
 struct Struct11I
-    x::Complex64
+    x::ComplexF32
 end
 
-function test_struct11{Struct}(::Type{Struct})
+function test_struct11(::Type{Struct}) where {Struct}
     a = Struct(0.8877077f0 + 0.4591081f0im)
     b = Float32(42)
 
@@ -428,15 +461,15 @@ test_struct11(Struct11)
 test_struct11(Struct11I)
 
 mutable struct Struct12
-    x::Complex64
-    y::Complex64
+    x::ComplexF32
+    y::ComplexF32
 end
 struct Struct12I
-    x::Complex64
-    y::Complex64
+    x::ComplexF32
+    y::ComplexF32
 end
 
-function test_struct12{Struct}(::Type{Struct})
+function test_struct12(::Type{Struct}) where {Struct}
     a = Struct(0.8877077f5 + 0.4591081f2im, 0.0004842868f0 - 6982.3265f3im)
     b = Float32(42)
 
@@ -453,13 +486,13 @@ test_struct12(Struct12)
 test_struct12(Struct12I)
 
 mutable struct Struct13
-    x::Complex128
+    x::ComplexF64
 end
 struct Struct13I
-    x::Complex128
+    x::ComplexF64
 end
 
-function test_struct13{Struct}(::Type{Struct})
+function test_struct13(::Type{Struct}) where {Struct}
     a = Struct(42968.97560380495 - 803.0576845153616im)
     b = Float64(42)
 
@@ -483,7 +516,7 @@ struct Struct14I
     y::Float32
 end
 
-function test_struct14{Struct}(::Type{Struct})
+function test_struct14(::Type{Struct}) where {Struct}
     a = Struct(0.024138331f0, 0.89759064f32)
     b = Float32(42)
 
@@ -508,7 +541,7 @@ struct Struct15I
     y::Float64
 end
 
-function test_struct15{Struct}(::Type{Struct})
+function test_struct15(::Type{Struct}) where {Struct}
     a = Struct(4.180997967273657, -0.404218594294923)
     b = Float64(42)
 
@@ -541,7 +574,7 @@ struct Struct16I
     c::Float64
 end
 
-function test_struct16{Struct}(::Type{Struct})
+function test_struct16(::Type{Struct}, quoteplz = false) where {Struct}
     a = Struct(0.1604656f0, 0.6297606f0, 0.83588994f0,
                0.6460273620993535, 0.9472692581106656, 0.47328535437352093)
     b = Float32(42)
@@ -549,7 +582,11 @@ function test_struct16{Struct}(::Type{Struct})
     if Struct === Struct16
         x = ccall((:test_16, libccalltest), Struct16, (Struct16, Float32), a, b)
     else
-        x = ccall((:test_16, libccalltest), Struct16I, (Struct16I, Float32), a, b)
+        if quoteplz
+          x = eval(:(ccall((:test_16, libccalltest), Struct16I, (Struct16I, Float32), $(QuoteNode(a)), Float32(42))))
+        else
+          x = ccall((:test_16, libccalltest), Struct16I, (Struct16I, Float32), a, b)
+        end
     end
 
     @test x.x ≈ a.x + b*1
@@ -559,8 +596,10 @@ function test_struct16{Struct}(::Type{Struct})
     @test x.b ≈ a.b + b*5
     @test x.c ≈ a.c - b*6
 end
-test_struct16(Struct16)
-test_struct16(Struct16I)
+
+test_struct16(Struct16, false)
+test_struct16(Struct16I, false)
+test_struct16(Struct16I, true)
 
 mutable struct Struct17
     a::Int8
@@ -571,7 +610,7 @@ struct Struct17I
     b::Int16
 end
 
-function test_struct17{Struct}(::Type{Struct})
+function test_struct17(::Type{Struct}) where {Struct}
     a = Struct(2, 10)
     b = Int8(2)
 
@@ -598,7 +637,7 @@ struct Struct18I
     c::Int8
 end
 
-function test_struct18{Struct}(::Type{Struct})
+function test_struct18(::Type{Struct}) where {Struct}
     a = Struct(2, 10, -3)
     b = Int8(2)
 
@@ -638,7 +677,7 @@ end
 copy(a::Struct_Big) = Struct_Big(a.x, a.y, a.z)
 copy(a::Struct_BigI) = a
 
-function test_struct_big{Struct}(::Type{Struct})
+function test_struct_big(::Type{Struct}) where {Struct}
     a = Struct(424,-5,Int8('Z'))
     a2 = copy(a)
 
@@ -710,7 +749,7 @@ function verify_huge(init, a, b)
     for i = 1:nfields(a)
         @test getfield(init, i) === getfield(a, i)
     end
-    # make sure b was modifed as expected
+    # make sure b was modified as expected
     a1, b1 = getfield(a, 1), getfield(b, 1)
     while isa(a1, Tuple)
         @test a1[2:end] === b1[2:end]
@@ -752,100 +791,200 @@ end
 ## cfunction roundtrip
 
 verbose && Libc.flush_cstdio()
+verbose && println("Testing cfunction closures: ")
+
+# helper Type for testing that constructors work
+# with cfucntion and that object identity is preserved
+mutable struct IdentityTestKV{K, V}
+    (T::Type{<:IdentityTestKV})(S) = (@test T === S; T)
+end
+
+@noinline function testclosure(f, a::T, permanent::Bool=false, tt::Type{S}=Any) where {T, S}
+    @nospecialize(f, a, tt)
+    # generic API 1
+    cf = @cfunction $f Ref{T} (Ref{T},)
+    GC.gc()
+    @test cf.ptr != C_NULL
+    @test cf.f === f
+    @test (cf._1 == C_NULL) == permanent
+    @test (cf._2 == C_NULL) == permanent
+    @assert cf === Base.cconvert(Ptr{Cvoid}, cf)
+    GC.@preserve cf begin
+        fptr = Base.unsafe_convert(Ptr{Cvoid}, cf)
+        b = ccall(fptr, Ref{T}, (Ref{T},), a)
+    end
+    # generic API 2
+    cf2 = @cfunction $f Any (Ref{S},)
+    GC.gc()
+    @test cf2.ptr != C_NULL
+    @test cf2.f === f
+    @test (cf2._1 == C_NULL) == permanent
+    @test (cf2._2 == C_NULL) == permanent
+    @assert cf2 === Base.cconvert(Ptr{Cvoid}, cf2)
+    GC.@preserve cf2 begin
+        fptr = Base.unsafe_convert(Ptr{Cvoid}, cf2)
+        b = ccall(fptr, Any, (Ref{S},), a)
+    end
+    return b
+end
+
+# We can't (currently) execute some of these signatures (without compile-all),
+# but we can at least look at some of the generated code
+function check_code_trampoline(f, t, n::Int)
+    @nospecialize(f, t)
+    @test Base.return_types(f, t) == Any[Any]
+    llvm = sprint(code_llvm, f, t)
+    @test count(x -> true, eachmatch(r"@jl_get_cfunction_trampoline\(", llvm)) == n
+end
+check_code_trampoline(testclosure, (Any, Any, Bool, Type), 2)
+check_code_trampoline(testclosure, (Any, Int, Bool, Type{Int}), 2)
+check_code_trampoline(testclosure, (Any, String, Bool, Type{String}), 2)
+check_code_trampoline(testclosure, (typeof(identity), Any, Bool, Type), 2)
+check_code_trampoline(testclosure, (typeof(identity), Int, Bool, Type{Int}), 0)
+check_code_trampoline(testclosure, (typeof(identity), String, Bool, Type{String}), 0)
+
+function g(i)
+    x = -332210 + i
+    y = "foo"
+    a(z) = x
+    b(z) = y
+    c(z) = (y = z)
+    IdentityTestVK{V, K} = IdentityTestKV{K, V}
+    @test IdentityTestVK !== IdentityTestKV
+    @test IdentityTestVK == IdentityTestKV
+    for _ = 1:5
+        @test testclosure(a, 23) == -332210 + i
+        @test testclosure(b, "bar") == "foo"
+        @test testclosure(c, "bar") == "bar"
+        @test testclosure(b, "foo") == "bar"
+        @test testclosure(c, "foo") == "foo"
+        @test testclosure(identity, IdentityTestKV, true) === IdentityTestKV
+        @test testclosure(identity, IdentityTestVK, true) === IdentityTestVK
+        @test testclosure(IdentityTestKV, IdentityTestKV, true) === IdentityTestKV
+        @test testclosure(IdentityTestVK, IdentityTestVK, false) === IdentityTestVK
+    end
+end
+g(1)
+g(2)
+g(3)
+
 verbose && println("Testing cfunction roundtrip: ")
 
 cf64 = 2.84+5.2im
 cf32 = 3.34f0+53.2f0im
-ci32 = Complex{Int32}(Int32(10),Int32(31))
-ci64 = Complex{Int64}(Int64(20),Int64(51))
+ci32 = Complex{Int32}(Int32(10), Int32(31))
+ci64 = Complex{Int64}(Int64(20), Int64(51))
 s1 = Struct1(352.39422f23, 19.287577)
-==(a::Struct1,b::Struct1) = a.x == b.x && a.y == b.y
+==(a::Struct1, b::Struct1) = (a.x == b.x && a.y == b.y)
 
-for (t,v) in ((Complex{Int32},:ci32),(Complex{Int64},:ci64),
-              (Complex64,:cf32),(Complex128,:cf64),(Struct1,:s1))
-    fname = Symbol("foo",v)
-    fname1 = Symbol("foo1",v)
+for (t, v) in ((Complex{Int32}, :ci32), (Complex{Int64}, :ci64),
+              (ComplexF32, :cf32), (ComplexF64, :cf64), (Struct1, :s1))
+    fname = Symbol("foo", v)
+    fname1 = Symbol("foo1", v)
+    a = copy(@eval $v)
+    verbose && println(t)
+    verbose && println("A: ", a)
     @eval begin
-        verbose && println($t)
-        a = copy($v)
-        verbose && println("A: ",a)
-        function $fname1(s::$t)
-            verbose && println("B: ",s)
+        global function $fname1(s::$t)
+            verbose && println("B: ", s)
             @test s == $v
-            @test s === a
+            @test s === $a
             global c = s
-            s
+            return s
         end
-        function $fname1(s)
+        global function $fname1(s)
             @assert false
         end
-        function $fname(s::$t)
-            verbose && println("B: ",s)
+        global function $fname(s::$t)
+            verbose && println("B: ", s)
             @test s == $v
             if($(t).mutable)
-                @test !(s === a)
+                @test !(s === $a)
             end
             global c = s
-            s
+            return s
         end
-        function $fname(s)
+        global function $fname(s)
             @assert false
         end
-        b = ccall(cfunction($fname1,Ref{$t},(Ref{$t},)),Ref{$t},(Ref{$t},),a)
-        verbose && println("C: ",b)
+    end
+    @eval let a = $a, b
+        b = testclosure($fname1, a, true)
+        verbose && println("C: ", b)
         @test b == $v
         @test b === a
         @test b === c
-        b = ccall(cfunction($fname,$t,($t,)),$t,($t,),a)
+        let cf = @cfunction($fname1, Ref{$t}, (Ref{$t},))
+            b = ccall(cf, Ref{$t}, (Ref{$t},), a)
+        end
+        verbose && println("C: ", b)
+        @test b == $v
+        @test b === a
+        @test b === c
+        let cf = @cfunction($fname, $t, ($t,))
+            b = ccall(cf, $t, ($t,), a)
+        end
         verbose && println("C: ",b)
         @test b == $v
         if ($(t).mutable)
             @test !(b === c)
             @test !(b === a)
         end
-        b = ccall(cfunction($fname1,$t,(Ref{$t},)),$t,(Ref{$t},),a)
+        let cf = @cfunction($fname1, $t, (Ref{$t},))
+            b = ccall(cf, $t, (Ref{$t},), a)
+        end
         verbose && println("C: ",b)
         @test b == $v
         if ($(t).mutable)
             @test !(b === c)
             @test !(b === a)
         end
-        b = ccall(cfunction($fname,Ref{$t},($t,)),Ref{$t},($t,),a)
+        let cf = @cfunction($fname, Ref{$t}, ($t,))
+            b = ccall(cf, Ref{$t}, ($t,), a)
+        end
         verbose && println("C: ",b)
         @test b == $v
         @test b === c
         if ($(t).mutable)
             @test !(b === a)
         end
-        b = ccall(cfunction($fname,Any,(Ref{$t},)),Any,(Ref{$t},),$v)
+        let cf = @cfunction($fname, Any, (Ref{$t},))
+            b = ccall(cf, Any, (Ref{$t},), $v)
+        end
         verbose && println("C: ",b)
         @test b == $v
         @test b === c
         if ($(t).mutable)
             @test !(b === a)
         end
-        b = ccall(cfunction($fname,Any,(Ref{Any},)),Any,(Ref{Any},),$v)
+        let cf = @cfunction($fname, Any, (Ref{Any},))
+            b = ccall(cf, Any, (Ref{Any},), $v)
+        end
         @test b == $v
         @test b === c
         if ($(t).mutable)
             @test !(b === a)
         end
-        @test_throws TypeError ccall(cfunction($fname,Ref{AbstractString},(Ref{Any},)),Any,(Ref{Any},),$v)
-        @test_throws TypeError ccall(cfunction($fname,AbstractString,(Ref{Any},)),Any,(Ref{Any},),$v)
+        let cf = @cfunction($fname, Ref{AbstractString}, (Ref{Any},))
+            @test_throws TypeError ccall(cf, Any, (Ref{Any},), $v)
+        end
+        let cf = @cfunction($fname, AbstractString, (Ref{Any},))
+            @test_throws TypeError ccall(cf, Any, (Ref{Any},), $v)
+        end
     end
 end
 
 # issue 13031
 foo13031(x) = Cint(1)
-foo13031p = cfunction(foo13031, Cint, (Ref{Tuple{}},))
+foo13031p = @cfunction(foo13031, Cint, (Ref{Tuple{}},))
 ccall(foo13031p, Cint, (Ref{Tuple{}},), ())
 
 foo13031(x,y,z) = z
-foo13031p = cfunction(foo13031, Cint, (Ref{Tuple{}},Ref{Tuple{}},Cint))
+foo13031p = @cfunction(foo13031, Cint, (Ref{Tuple{}}, Ref{Tuple{}}, Cint))
 ccall(foo13031p, Cint, (Ref{Tuple{}},Ref{Tuple{}},Cint), (), (), 8)
 
 # issue 17219
-function ccall_reassigned_ptr(ptr::Ptr{Void})
+function ccall_reassigned_ptr(ptr::Ptr{Cvoid})
     ptr = Libdl.dlsym(Libdl.dlopen(libccalltest), "test_echo_p")
     ccall(ptr, Any, (Any,), "foo")
 end
@@ -878,17 +1017,31 @@ end
 
 @test ccall(:jl_getpagesize, Clong, ()) == @threadcall(:jl_getpagesize, Clong, ())
 
+# make sure our malloc/realloc/free adapters are thread-safe and repeatable
+for i = 1:8
+    ptr = @threadcall(:jl_malloc, Ptr{Cint}, (Csize_t,), sizeof(Cint))
+    @test ptr != C_NULL
+    unsafe_store!(ptr, 3)
+    @test unsafe_load(ptr) == 3
+    ptr = @threadcall(:jl_realloc, Ptr{Cint}, (Ptr{Cint}, Csize_t,), ptr, 2 * sizeof(Cint))
+    @test ptr != C_NULL
+    unsafe_store!(ptr, 4, 2)
+    @test unsafe_load(ptr, 1) == 3
+    @test unsafe_load(ptr, 2) == 4
+    @threadcall(:jl_free, Cvoid, (Ptr{Cint},), ptr)
+end
+
 # Pointer finalizer (issue #15408)
 let A = [1]
-    ccall((:set_c_int, libccalltest), Void, (Cint,), 1)
+    ccall((:set_c_int, libccalltest), Cvoid, (Cint,), 1)
     @test ccall((:get_c_int, libccalltest), Cint, ()) == 1
-    finalizer(A, cglobal((:finalizer_cptr, libccalltest), Void))
+    finalizer(cglobal((:finalizer_cptr, libccalltest), Cvoid), A)
     finalize(A)
     @test ccall((:get_c_int, libccalltest), Cint, ()) == -1
 end
 
 # Pointer finalizer at exit (PR #19911)
-let result = readstring(`$(Base.julia_cmd()) --startup-file=no -e "A = Ref{Cint}(42); finalizer(A, cglobal((:c_exit_finalizer, \"$libccalltest\"), Void))"`)
+let result = read(`$(Base.julia_cmd()) --startup-file=no -e "A = Ref{Cint}(42); finalizer(cglobal((:c_exit_finalizer, \"$libccalltest\"), Cvoid), A)"`, String)
     @test result == "c_exit_finalizer: 42, 0"
 end
 
@@ -993,10 +1146,9 @@ if Sys.ARCH === :x86_64
         T = NTuple{4, VecElement{s}}
         @eval function rt_sse(a1::$T, a2::$T, a3::$T, a4::$T)
             return ccall(
-                cfunction(foo_ams, $T, ($T, $T, $T, $T)),
-                $T,
-                ($T, $T, $T, $T),
-                a1,  a2,  a3, a4)
+                @cfunction(foo_ams, $T, ($T, $T, $T, $T)),
+                $T, ($T, $T, $T, $T),
+                     a1, a2, a3, a4)
         end
 
         a1 = VecReg(ntuple(i -> VecElement(s(1i)), 4))
@@ -1131,7 +1283,7 @@ elseif Sys.ARCH === :powerpc64le || Sys.ARCH === :ppc64le
     (1024, 1023, 1022, 1021), (1025, 1024, 1023, 1022), (1026, 1025, 1024, 1023), (1027, 1026, 1025, 1024), (10028, 10027, 10026, 10025))
 
 elseif Sys.ARCH !== :i686 && Sys.ARCH !== :arm # TODO
-warn("ccall: no VecReg tests run for this platform")
+@warn "ccall: no VecReg tests run for this platform"
 
 end
 
@@ -1143,7 +1295,7 @@ function f17204(a)
     end
     return b
 end
-@test ccall(cfunction(f17204, Vector{Any}, Tuple{Vector{Any}}),
+@test ccall(@cfunction(f17204, Vector{Any}, (Vector{Any},)),
             Vector{Any}, (Vector{Any},), Any[1:10;]) == Any[11:20;]
 
 # This used to trigger incorrect ccall callee inlining.
@@ -1151,9 +1303,9 @@ end
 # Do not put these in a function.
 @noinline g17413() = rand()
 @inline f17413() = (g17413(); g17413())
-ccall((:test_echo_p, libccalltest), Ptr{Void}, (Any,), f17413())
+ccall((:test_echo_p, libccalltest), Ptr{Cvoid}, (Any,), f17413())
 for i in 1:3
-    ccall((:test_echo_p, libccalltest), Ptr{Void}, (Any,), f17413())
+    ccall((:test_echo_p, libccalltest), Ptr{Cvoid}, (Any,), f17413())
 end
 
 struct SpillPint
@@ -1238,32 +1390,40 @@ for i in 1:100
 end
 
 # issue #20835
-@test_throws ErrorException eval(:(f20835(x) = ccall(:fn, Void, (Ptr{typeof(x)},), x)))
-@test_throws UndefVarError  eval(:(f20835(x) = ccall(:fn, Something_not_defined_20835, (Ptr{typeof(x)},), x)))
+@test_throws(ErrorException("could not evaluate ccall argument type (it might depend on a local variable)"),
+             eval(:(f20835(x) = ccall(:fn, Cvoid, (Ptr{typeof(x)},), x))))
+@test_throws(UndefVarError(:Something_not_defined_20835),
+             eval(:(f20835(x) = ccall(:fn, Something_not_defined_20835, (Ptr{typeof(x)},), x))))
 
-@noinline f21104at(::Type{T}) where {T} = ccall(:fn, Void, (Nullable{T},), 0)
-@noinline f21104rt(::Type{T}) where {T} = ccall(:fn, Nullable{T}, ())
-@test code_llvm(DevNull, f21104at, (Type{Float64},)) === nothing
-@test code_llvm(DevNull, f21104rt, (Type{Float64},)) === nothing
-@test try
-          f21104at(Float64)
-          "unreachable"
-      catch ex
-          (ex::ErrorException).msg
-      end == "ccall: the type of argument 1 doesn't correspond to a C type"
-@test try
-          f21104rt(Float64)
-          "unreachable"
-      catch ex
-          (ex::ErrorException).msg
-      end == "ccall: return type doesn't correspond to a C type"
+@noinline f21104at(::Type{T}) where {T} = ccall(:fn, Cvoid, (Some{T},), Some(0))
+@noinline f21104rt(::Type{T}) where {T} = ccall(:fn, Some{T}, ())
+@test code_llvm(devnull, f21104at, (Type{Float64},)) === nothing
+@test code_llvm(devnull, f21104rt, (Type{Float64},)) === nothing
+@test_throws(ErrorException("ccall argument 1 doesn't correspond to a C type"),
+             f21104at(Float64))
+@test_throws(ErrorException("ccall return type doesn't correspond to a C type"),
+             f21104rt(Float64))
 
+# test for malformed syntax errors
+@test Expr(:error, "more arguments than types for ccall") == Meta.lower(@__MODULE__, :(ccall(:fn, A, (), x)))
+@test Expr(:error, "more arguments than types for ccall") == Meta.lower(@__MODULE__, :(ccall(:fn, A, (B,), x, y)))
+@test Expr(:error, "more arguments than types for ccall") == Meta.lower(@__MODULE__, :(ccall(:fn, A, (B,), x, y, z)))
+@test Expr(:error, "more arguments than types for ccall") == Meta.lower(@__MODULE__, :(ccall(:fn, A, (B,), x, y)))
+@test Expr(:error, "more arguments than types for ccall") == Meta.lower(@__MODULE__, :(ccall(:fn, A, (B,), x, y, z)))
+@test Expr(:error, "more arguments than types for ccall") == Meta.lower(@__MODULE__, :(ccall(:fn, A, (B, C), x, y, z)))
+@test Expr(:error, "more types than arguments for ccall") == Meta.lower(@__MODULE__, :(ccall(:fn, A, (B,),)))
+@test Expr(:error, "more types than arguments for ccall") == Meta.lower(@__MODULE__, :(ccall(:fn, A, (B, C), )))
+@test Expr(:error, "more types than arguments for ccall") == Meta.lower(@__MODULE__, :(ccall(:fn, A, (B..., C...), )))
+@test Expr(:error, "C ABI prohibits vararg without one required argument") == Meta.lower(@__MODULE__, :(ccall(:fn, A, (B...,), x)))
+@test Expr(:error, "only the trailing ccall argument type should have \"...\"") == Meta.lower(@__MODULE__, :(ccall(:fn, A, (A, B..., C...), a, x)))
+@test Expr(:error, "only the trailing ccall argument type should have \"...\"") == Meta.lower(@__MODULE__, :(ccall(:fn, A, (A, B..., C...), a, x, y, z)))
+@test Expr(:error, "more types than arguments for ccall") == Meta.lower(@__MODULE__, :(ccall(:fn, A, (B, C...), )))
 
 # cfunction on non-function singleton
 struct CallableSingleton
 end
 (::CallableSingleton)(x, y) = x + y
-@test ccall(cfunction(CallableSingleton(), Int, Tuple{Int,Int}),
+@test ccall(@cfunction(CallableSingleton(), Int, (Int, Int)),
             Int, (Int, Int), 1, 2) === 3
 
 # 19805
@@ -1271,9 +1431,107 @@ mutable struct callinfos_19805{FUNC_FT<:Function}
     f :: FUNC_FT
 end
 
-evalf_callback_19805{FUNC_FT}(ci::callinfos_19805{FUNC_FT}) = ci.f(0.5)::Float64
+evalf_callback_19805(ci::callinfos_19805{FUNC_FT}) where {FUNC_FT} = ci.f(0.5)::Float64
 
-evalf_callback_c_19805{FUNC_FT}(ci::callinfos_19805{FUNC_FT}) = cfunction(
+evalf_callback_c_19805(ci::callinfos_19805{FUNC_FT}) where {FUNC_FT} = @cfunction(
     evalf_callback_19805, Float64, (callinfos_19805{FUNC_FT},))
 
-@test_throws ErrorException evalf_callback_c_19805( callinfos_19805(sin) )
+@test_throws(ErrorException("cfunction argument 1 doesn't correspond to a C type"),
+             evalf_callback_c_19805( callinfos_19805(sin) ))
+@test_throws(ErrorException("cfunction argument 2 doesn't correspond to a C type"),
+             @cfunction(+, Int, (Int, Nothing)))
+@test_throws(ErrorException("cfunction: Vararg syntax not allowed for argument list"),
+             @cfunction(+, Int, (Vararg{Int},)))
+@test_throws(ErrorException("could not evaluate cfunction argument type (it might depend on a local variable)"),
+             @eval () -> @cfunction(+, Int, (Ref{T}, Ref{T})) where T)
+@test_throws(ErrorException("could not evaluate cfunction return type (it might depend on a local variable)"),
+             @eval () -> @cfunction(+, Ref{T}, (Int, Int)) where T)
+@test_throws(ErrorException("cfunction return type Ref{Any} is invalid. Use Any or Ptr{Any} instead."),
+             @cfunction(+, Ref{Any}, (Int, Int)))
+
+# test Ref{abstract_type} calling parameter passes a heap box
+abstract type Abstract22734 end
+struct Bits22734 <: Abstract22734
+    x::Int
+    y::Float64
+end
+function cb22734(ptr::Ptr{Cvoid})
+    GC.gc()
+    obj = unsafe_pointer_to_objref(ptr)::Bits22734
+    obj.x + obj.y
+end
+ptr22734 = @cfunction(cb22734, Float64, (Ptr{Cvoid},))
+function caller22734(ptr)
+    obj = Bits22734(12, 20)
+    ccall(ptr, Float64, (Ref{Abstract22734},), obj)
+end
+@test caller22734(ptr22734) === 32.0
+
+# 26297#issuecomment-371165725
+#   test that the first argument to cglobal is recognized as a tuple literal even through
+#   macro expansion
+macro cglobal26297(sym)
+    :(cglobal(($(esc(sym)), libccalltest), Cint))
+end
+cglobal26297() = @cglobal26297(:global_var)
+@test cglobal26297() != C_NULL
+
+# issue #26607
+noop_func_26607 = () -> nothing
+function callthis_26607(args)
+    @cfunction(noop_func_26607, Cvoid, ())
+    return nothing
+end
+@test callthis_26607(Int64(0)) === nothing
+@test callthis_26607(Int32(0)) === nothing
+
+# issue #27178 (cfunction special case in inlining)
+mutable struct CallThisFunc27178{FCN_TYPE}
+    fcn::FCN_TYPE
+end
+
+callback27178(cb::CTF) where CTF<:CallThisFunc27178 = nothing
+@inline make_cfunc27178(cbi::CI) where CI = @cfunction(callback27178, Cvoid, (Ref{CI},))
+get_c_func(fcn::FCN_TYPE) where {FCN_TYPE<:Function} = return make_cfunc27178(CallThisFunc27178(fcn))
+@test isa(get_c_func(sin), Ptr)
+
+# issue #27215
+function once_removed()
+    function mycompare(a, b)::Cint
+        return (a < b) ? -1 : ((a > b) ? +1 : 0)
+    end
+    mycompare_c = @cfunction($mycompare, Cint, (Ref{Cdouble}, Ref{Cdouble}))
+end
+@test isa(once_removed(), Base.CFunction)
+
+# issue #27478
+function ccall27478()
+    module_lib = Libdl.dlopen("libjulia")
+    ccall(Libdl.dlsym(module_lib, "getpid"), Cint, ())
+end
+@test code_typed(ccall27478, ()) isa Array
+
+# issue #27477
+@eval module Pkg27477
+const libccalltest = $libccalltest
+end
+
+module Test27477
+using ..Pkg27477
+test27477() = ccall((:ctest, Pkg27477.libccalltest), Complex{Int}, (Complex{Int},), 1 + 2im)
+end
+
+@test Test27477.test27477() == 2 + 0im
+
+# issue #31073
+let
+    a = ['0']
+    arr = Vector{Char}(undef, 2)
+    ptr = pointer(arr)
+    elsz = sizeof(Char)
+    na = length(a)
+    nba = na * elsz
+    ptr = eval(:(ccall(:memcpy, Ptr{Cvoid}, (Ptr{Cvoid}, Ptr{Cvoid}, UInt), $(arr), $(a), $(nba))))
+    @test isa(ptr, Ptr{Cvoid})
+    @test arr[1] == '0'
+end
